LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;

LIBRARY topic;
USE topic.topic_tb_pkg.ALL;

ENTITY clock_reset_manager IS
   GENERIC(
      p_number_of_output_clocks : natural := 1;
      p_clk_periods             : t_time_array;
      p_enable_gsr              : boolean := true;
      p_gsr_duration            : time    := 100 ns
   );
   PORT(
      o_gsr         : OUT std_logic;
      o_clks        : OUT std_logic_vector(p_number_of_output_clocks - 1 DOWNTO 0);
      o_rsts        : OUT std_logic_vector(p_number_of_output_clocks - 1 DOWNTO 0);
      o_rsts_n      : OUT std_logic_vector(p_number_of_output_clocks - 1 DOWNTO 0);
      o_resets_done : OUT std_logic;
      i_stop_clocks : IN  std_logic
   );
END ENTITY;

ARCHITECTURE RTL OF clock_reset_manager IS

   SIGNAL n_gsr         : std_logic                                                := '1';
   SIGNAL n_clks        : std_logic_vector(p_number_of_output_clocks - 1 DOWNTO 0) := (OTHERS => '0');
   SIGNAL n_rsts        : std_logic_vector(p_number_of_output_clocks - 1 DOWNTO 0) := (OTHERS => '1');
   SIGNAL n_resets_done : std_logic_vector(p_number_of_output_clocks - 1 DOWNTO 0) := (OTHERS => '0');

BEGIN

   o_gsr         <= n_gsr;
   o_clks        <= n_clks;
   o_rsts        <= n_rsts;
   o_rsts_n      <= NOT n_rsts;
   o_resets_done <= '1' WHEN n_resets_done = (n_resets_done'RANGE => '1') ELSE '0';

   PROCESS
   BEGIN
      IF (p_enable_gsr) THEN
         WAIT FOR p_gsr_duration;
      END IF;
      n_gsr <= '0';
      WAIT;
   END PROCESS;

   clk_rst_gen : FOR i IN 0 TO p_number_of_output_clocks - 1 GENERATE
   BEGIN
      -- Start generating clock after GSR (Global Set Reset) is done
      n_clks(i) <= NOT n_clks(i) AFTER p_clk_periods(i) / 2 WHEN n_gsr = '0' AND i_stop_clocks = '0' ELSE '0';

      -- Generate reset for this clock domain
      PROCESS
      BEGIN
         n_rsts(i) <= '1';

         FOR j IN 0 TO 4 LOOP
            WAIT UNTIL rising_edge(n_clks(i));
         END LOOP;

         n_rsts(i)        <= '0';
         WAIT UNTIL rising_edge(n_clks(i));
         n_resets_done(i) <= '1';
         WAIT;
      END PROCESS;

   END GENERATE;

END ARCHITECTURE RTL;

