-- File: dyplo_uart.vhd
--
-- (c) Copyright 2012 - 2014 Topic Embedded Systems. All rights reserved.
--
-- This file contains confidential and proprietary information of Topic
-- Embedded Systems and is protected under U.S. and international copyright
-- and other intellectual property laws.
--
-- DISCLAIMER
-- This disclaimer is not a license and does not grant any rights to the
-- materials distributed herewith. Except as otherwise provided in a valid
-- license issued to you by Topic Embedded Systems, and to the maximum extent
-- permitted by applicable law:
-- (1) THESE MATERIALS ARE MADE AVAILABLE "AS IS" AND WITH ALL FAULTS,
--     AND TOPIC EMBEDDED SYSTEMS HEREBY DISCLAIMS ALL WARRANTIES AND
--     CONDITIONS, EXPRESS, IMPLIED, OR STATUTORY, INCLUDING BUT NOT LIMITED
--     TO WARRANTIES OF MERCHANTABILITY, NON-INFRINGEMENT, OR FITNESS FOR ANY
--     PARTICULAR PURPOSE;
-- and
-- (2) Topic Embedded Systems shall not be liable (whether in contract or tort,
--     including negligence, or under any other theory of liability) for any
--     loss or damage of any kind or nature related to, arising under or in
--     connection with these materials, including for any direct, or any
--     indirect, special, incidental, or consequential loss or damage
--     (including loss of data, profits, goodwill, or any type of loss or
--     damage suffered as a result of any action brought by a third party)
--     even if such damage or loss was reasonably foreseeable or Topic Embedded
--     Systems had been advised of the possibility of the same.
--
-- CRITICAL APPLICATIONS
-- Topic Embedded System products are not designed or intended to be fail-safe,
-- or for use in any application requiring fail-safe performance, such as
-- life-support or safety devices or systems, Class III medical devices,
-- nuclear facilities, applications related to the deployment of airbags, or
-- any other applications that could lead to death, personal injury, or severe
-- property or environmental damage (individually and collectively, "Critical
-- Applications"). Customer assumes the sole risk and liability of any use of
-- Topic Embedded Systems products in Critical Applications, subject only to
-- applicable laws and regulations governing limitations on product liability.
--
-- THIS COPYRIGHT NOTICE AND DISCLAIMER MUST BE RETAINED AS PART OF THIS FILE
-- AT ALL TIMES.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity dyplo_uart is
   port(
      -- Miscellaneous
      clk        : in  std_logic;
      rst        : in  std_logic;
      rst_ack    : out std_logic;
      en         : in  std_logic;
      -- Processor interface bus
      -- Streaming data to serial port TxD channel
      txd_tvalid : in  std_logic;
      txd_tready : out std_logic;
      txd_tdata  : in  std_logic_vector(31 downto 0);
      txd        : out std_logic;
      -- Streaming data from FPGA fabric to Dyplo block
      rxd_tvalid : out std_logic;
      rxd_tready : in  std_logic;
      rxd_tdata  : out std_logic_vector(31 downto 0);
      rxd        : in  std_logic
   );
end dyplo_uart;

architecture rtl of dyplo_uart is
   constant BAUDRATE_DIVISION   : unsigned(15 downto 0) := to_unsigned(216, 16); -- (100MHz / 460800 baud) - 1;
   constant SAMPLERATE_DIVISION : unsigned(23 downto 0) := to_unsigned(1236951, 24); -- ((460800 baud * 16)/100MHz) * 2^24;
   constant SM_WORD_MODE        : std_logic             := '1';

   type rxd_table_type is array (0 to 31) of std_logic_vector(1 downto 0);
   constant rxd_table : rxd_table_type := ("00", -- "00000", msb = bit level, lsb = start bit level
                                           "00", -- "00001"
                                           "00", -- "00010"
                                           "00", -- "00011"
                                           "00", -- "00100"
                                           "00", -- "00101"
                                           "00", -- "00110"
                                           "10", -- "00111"
                                           "00", -- "01000"
                                           "00", -- "01001"
                                           "00", -- "01010"
                                           "10", -- "01011"
                                           "00", -- "01100"
                                           "10", -- "01101"
                                           "10", -- "01110"
                                           "10", -- "01111"
                                           "01", -- "10000"
                                           "01", -- "10001"
                                           "01", -- "10010"
                                           "10", -- "10011"
                                           "01", -- "10100"
                                           "10", -- "10101"
                                           "10", -- "10110"
                                           "10", -- "10111"
                                           "01", -- "11000"
                                           "11", -- "11001"
                                           "11", -- "11010"
                                           "10", -- "11011"
                                           "10", -- "11100"
                                           "10", -- "11101"
                                           "10", -- "11110"
                                           "10"); -- "11111"
   attribute ram_style : string;
   attribute ram_style of rxd_table : constant is "DISTRIBUTED";

   type sm_transmit_states is (TXD_FIFO_FETCH, TXD_PREPARE, TXD_TRANSMIT);
   signal sm_transmit  : sm_transmit_states;
   signal txd_time_cnt : unsigned(15 downto 0);
   signal txd_bit_cnt  : unsigned(3 downto 0);
   signal txd_word_cnt : unsigned(1 downto 0);
   signal txd_shft_reg : std_logic_vector(10 downto 0);
   signal txd_data_reg : std_logic_vector(31 downto 0);
   signal txd_tready_i : std_logic;

   type sm_receive_states is (RXD_START, RXD_RECEIVE);
   signal sm_receive   : sm_receive_states;
   signal rxd_bit_cnt  : unsigned(3 downto 0);
   signal rxd_word_cnt : unsigned(1 downto 0);
   signal rxd_shft_reg : std_logic_vector(4 downto 0);

   signal rxd_start_event  : std_logic;
   signal rxd_bit_level    : std_logic;
   signal rxd_table_value  : std_logic_vector(1 downto 0);
   signal rxd_sample_accu  : unsigned(24 downto 0);
   signal rxd_sample_event : std_logic;
   signal rxd_sample_reg   : std_logic_vector(7 downto 0);
   signal rxd_meta         : std_logic_vector(2 downto 0);
   signal rxd_tvalid_i     : std_logic := '0';
   signal rxd_sample_cnt   : unsigned(3 downto 0);
   signal rxd_valid        : std_logic;
   signal rxd_valid_p1     : std_logic;
   signal rxd_data         : std_logic_vector(7 downto 0);
   signal rxd_fifo_error   : unsigned(7 downto 0);

begin
   rst_ack_p : process(clk)
   begin
      if (rising_edge(clk)) then
         rst_ack <= rst;
      end if;
   end process rst_ack_p;

   transmit : process(clk)
   begin
      if (rising_edge(clk)) then
         if (rst = '1') then
            txd_tready_i <= '0';
            sm_transmit  <= TXD_FIFO_FETCH;
            txd_shft_reg <= (others => '1');
            txd          <= '1';
         elsif (en = '1') then
            case sm_transmit is
               --------------------------------------------------------------
               when TXD_FIFO_FETCH =>
                  if (txd_tvalid = '1') then
                     txd_tready_i <= '1';
                     txd_data_reg <= txd_tdata;
                     if (SM_WORD_MODE = '1') then
                        txd_word_cnt <= "11";
                     else
                        txd_word_cnt <= "00";
                     end if;
                     sm_transmit <= TXD_PREPARE;
                  end if;
               --------------------------------------------------------------
               when TXD_PREPARE =>
                  txd_tready_i <= '0';
                  txd_shft_reg <= '0' & txd_data_reg(7 downto 0) & "11";
                  txd_data_reg <= X"FF" & txd_data_reg(31 downto 8);
                  txd_time_cnt <= BAUDRATE_DIVISION;
                  txd_bit_cnt  <= to_unsigned(10, 4);
                  sm_transmit  <= TXD_TRANSMIT;
               --------------------------------------------------------------
               when TXD_TRANSMIT =>
                  if (txd_time_cnt = 0) then
                     txd_time_cnt <= BAUDRATE_DIVISION;
                     txd_shft_reg <= txd_shft_reg(9 downto 0) & '1';
                     if (txd_bit_cnt = 1) then
                        if (txd_word_cnt = "00") then
                           sm_transmit <= TXD_FIFO_FETCH;
                        else
                           sm_transmit  <= TXD_PREPARE;
                           txd_word_cnt <= txd_word_cnt - 1;
                        end if;
                     else
                        txd_bit_cnt <= txd_bit_cnt - 1;
                     end if;
                  else
                     txd_time_cnt <= txd_time_cnt - 1;
                  end if;
            --------------------------------------------------------------
            end case;

            txd <= txd_shft_reg(10);

         end if;
      end if;
   end process transmit;

   rom_process : process(clk) is
   begin
      if (rising_edge(clk)) then
         rxd_table_value <= rxd_table(to_integer(unsigned(rxd_shft_reg)));
      end if;
   end process rom_process;

   rxd_start_event <= rxd_table_value(0);
   rxd_bit_level   <= rxd_table_value(1);

   receive : process(clk)
      variable rxd_sample_accu_var : unsigned(24 downto 0);
   begin
      if (rising_edge(clk)) then
         if (rst = '1') then
            rxd_sample_accu  <= (others => '0');
            rxd_sample_event <= '0';
            rxd_meta         <= (others => '1');
            rxd_tvalid_i     <= '0';
            sm_receive       <= RXD_START;
            rxd_sample_cnt   <= (others => '0');
            rxd_bit_cnt      <= (others => '0');
            rxd_word_cnt     <= (others => '0');
            rxd_shft_reg     <= (others => '1');
            rxd_valid        <= '1';
            rxd_fifo_error   <= (others => '0');
         elsif (en = '1') then

            -- Metastability avoidance
            rxd_meta <= rxd_meta(1 downto 0) & rxd;

            -- Sample event determination
            rxd_sample_accu_var := rxd_sample_accu + unsigned(('0' & SAMPLERATE_DIVISION));
            rxd_sample_accu     <= '0' & rxd_sample_accu_var(23 downto 0);
            rxd_sample_event    <= rxd_sample_accu_var(24);

            -- Sample serial line: 16x oversampling. The shift register content
            -- is applied to a lookup table to determine the start edge or the
            -- actual bit value.
            if (rxd_sample_event = '1') then
               rxd_shft_reg <= rxd_shft_reg(3 downto 0) & rxd_meta(2);
            end if;

            -- Based on the state of the lookup table values the UART protocol
            -- is interpreted.
            if (rxd_sample_event = '1') then
               case sm_receive is
                  ------------------------------------------------------------
                  when RXD_START =>
                     rxd_valid <= '0';
                     if (rxd_start_event = '1') then
                        rxd_sample_cnt <= to_unsigned(7, 4);
                        rxd_bit_cnt    <= to_unsigned(10, 4);
                        sm_receive     <= RXD_RECEIVE;
                     end if;
                  ------------------------------------------------------------
                  when RXD_RECEIVE =>
                     if (rxd_sample_cnt = 0) then
                        rxd_sample_cnt <= to_unsigned(15, 4);
                        rxd_sample_reg <= rxd_sample_reg(6 downto 0) & rxd_bit_level;
                        if (rxd_bit_cnt = 1) then
                           sm_receive <= RXD_START;
                           rxd_valid  <= '1';
                           rxd_data   <= rxd_sample_reg;
                        else
                           rxd_valid <= '0';
                        end if;
                        if (rxd_bit_cnt = 0) then
                           sm_receive <= RXD_START;
                        else
                           rxd_bit_cnt <= rxd_bit_cnt - 1;
                        end if;
                     else
                        rxd_sample_cnt <= rxd_sample_cnt - 1;
                     end if;
               ------------------------------------------------------------
               end case;
            end if;

            rxd_valid_p1 <= rxd_valid;
            if (rxd_tvalid_i = '1') and (rxd_tready = '1') then
               rxd_tvalid_i <= '0';
            elsif (rxd_valid = '1') and (rxd_valid_p1 = '0') then
               if (SM_WORD_MODE = '1') then
                  case rxd_word_cnt is
                     when "00"   => rxd_tdata(7 downto 0) <= rxd_data;
                     when "01"   => rxd_tdata(15 downto 8) <= rxd_data;
                     when "10"   => rxd_tdata(23 downto 16) <= rxd_data;
                     when "11"   => rxd_tdata(31 downto 24) <= rxd_data;
                     when others => null;
                  end case;
                  rxd_word_cnt <= rxd_word_cnt + 1;
                  if (rxd_word_cnt = "11") then
                     rxd_tvalid_i <= '1';
                  end if;
               else
                  rxd_tdata    <= X"000000" & rxd_data;
                  rxd_tvalid_i <= '1';
               end if;
            end if;

            if (rxd_tvalid_i = '1') and (rxd_valid = '1') and (rxd_valid_p1 = '0') then
               rxd_fifo_error <= rxd_fifo_error + 1;
            end if;

         end if;
      end if;
   end process receive;

   rxd_tvalid <= rxd_tvalid_i;
   txd_tready <= txd_tready_i;

end rtl;
