library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use std.textio.all;

package tb_env_pkg is

  file log_file             : text;  -- main log file

  shared variable v_log_file_status  : file_open_status;                     --file open status log file
  shared variable v_log_line         : line;                                 --store logging line for log file

  constant WORD_SIZE        : integer := 32;
  constant CMD_WORD_SIZE    : integer := WORD_SIZE;
  constant CMD_WORD_DEPTH   : integer := 16;

  constant LOG_FILE_DIR   : string := ("../../../../../simulation/test_framework/vhdl_sim/Log");
  constant RESULTS_LOG_FILE    		: string := (LOG_FILE_DIR & "/results.log");

  constant NO_FIFOS         : integer := 16;

  constant c_read_conf_no_check       : std_logic_vector(3 downto 0) := "0001"; --read addr
  constant c_read_conf_check_result   : std_logic_vector(3 downto 0) := "0010"; --read addr exp_data
  constant c_read_data                : std_logic_vector(3 downto 0) := "0011"; --read data
  constant c_req_irq_info             : std_logic_vector(3 downto 0) := "0100"; --request interrupt info
  constant c_req_read_fifo_status     : std_logic_vector(3 downto 0) := "0101"; --request read fifo status
  constant c_req_write_fifo_status    : std_logic_vector(3 downto 0) := "0110"; --request write fifo status
  constant c_write_conf               : std_logic_vector(3 downto 0) := "1000"; --write addr
  constant c_write_data               : std_logic_vector(3 downto 0) := "1001"; --write data
  constant c_clear_read_irq           : std_logic_vector(3 downto 0) := "1010"; --clear read irq
  constant c_clear_write_irq          : std_logic_vector(3 downto 0) := "1011"; --clear write irq
  constant c_set_read_threshold       : std_logic_vector(3 downto 0) := "1100"; --set read threshold
  constant c_set_write_threshold      : std_logic_vector(3 downto 0) := "1101"; --set write threshold
  constant c_set_read_irq             : std_logic_vector(3 downto 0) := "1110"; --set read irq
  constant c_set_write_irq            : std_logic_vector(3 downto 0) := "1111"; --set write irq

  constant c_cpu_read_fifo_size     : integer := 256;
  constant c_cpu_write_fifo_size    : integer := 256;

  type fifo_cmd_type is array (NO_FIFOS - 1 downto 0) of std_logic_vector(99 downto 0);

  subtype word_subtype is string(1 to WORD_SIZE);
  type word_type is
    array (natural range <> ) of word_subtype;

  type size_type is
    array (natural range <> ) of integer;

  type cmd_record is
    record
      stim_file : string(1 to 64);
      line_cnt  : integer;
      word      : word_type(CMD_WORD_DEPTH-1 downto 0);
      size      : size_type(CMD_WORD_DEPTH-1 downto 0);
      cnt       : integer;
      valid     : boolean;
    end record cmd_record;

  function read_time(input_string: in string)
    return time;

  function astr_to_slv (s: in string)
    return std_logic_vector;

  function hstr_to_slv (s: in string)
    return std_logic_vector;

  function char_to_sl (c: in character)
    return std_logic;

  function bstr_to_slv (s: in string)
    return std_logic_vector;

  procedure proc_str_to_int (
    variable str        : in  string;
    variable int        : out integer;
    variable result     : out boolean
  );

  function func_lowercase (word : in word_subtype)
    return word_subtype;

  procedure proc_split_line (
    variable line_buf       : in  line;
    variable word_array     : out word_type;
    variable size_array     : out size_type;
    variable word_cnt       : out integer;
    variable empty_line     : out boolean;
    variable comment_line   : out boolean
  );

  procedure proc_get_value (
    variable str            : in  string;
    variable slv            : out std_logic_vector;
    variable result         : out boolean;
    variable len            : out integer
  );

  function to_hstring (value : std_logic_vector)
    return string;

  function to_intstring (value : std_logic_vector)
    return string;

  function int_to_string( input : integer )
    return string;

  procedure str_read(file in_file: TEXT;
                     res_string: out string);

end package tb_env_pkg;


package body tb_env_pkg is

--read_time
--This function convert a string to time.
function read_time(input_string: in string)
  return time is

  variable v_temp_line  : line;    --store line data
  variable v_delay      : time;    --delay time
  variable v_result     : boolean; --result indicator
begin
    write(v_temp_line,input_string);
    read(v_temp_line, v_delay, v_result);
    if (not(v_result)) then
      report"ERROR function read_time"
        severity failure;
    else
      return v_delay;
    end if;
end function read_time;


function astr_to_slv (s: in string)
  return std_logic_vector is

  variable slv: std_logic_vector((s'high-s'low)*8+7 downto 0);
  variable k: integer;  -- position in std_logic_vector
begin
  k := 0;
  for i in s'high downto s'low loop
    slv(k+7 downto k) := std_logic_vector(to_unsigned(character'pos(s(i)),8));
    k := k + 8;
  end loop;
  return slv;
end function astr_to_slv;

function hstr_to_slv (s: in string)
  return std_logic_vector is

  variable slv: std_logic_vector((s'high-s'low)*4+3 downto 0);
  variable k: integer;  -- position in std_logic_vector
  variable j: integer;  -- to filter underscores
  variable u: integer;  -- help for position in std_logic_vector
  variable l: integer;  -- help for position in std_logic_vector
begin
  k := 0;
  j := 0;  -- tofilter underscores
  for i in s'high downto s'low loop
    u := k + 3 - j;
    l := k - j;
    case s(i) is
      when '0'       => slv(u downto l) := "0000";
      when '1'       => slv(u downto l) := "0001";
      when '2'       => slv(u downto l) := "0010";
      when '3'       => slv(u downto l) := "0011";
      when '4'       => slv(u downto l) := "0100";
      when '5'       => slv(u downto l) := "0101";
      when '6'       => slv(u downto l) := "0110";
      when '7'       => slv(u downto l) := "0111";
      when '8'       => slv(u downto l) := "1000";
      when '9'       => slv(u downto l) := "1001";
      when 'a' | 'A' => slv(u downto l) := "1010";
      when 'b' | 'B' => slv(u downto l) := "1011";
      when 'c' | 'C' => slv(u downto l) := "1100";
      when 'd' | 'D' => slv(u downto l) := "1101";
      when 'e' | 'E' => slv(u downto l) := "1110";
      when 'f' | 'F' => slv(u downto l) := "1111";
      when 'z' | 'Z' => slv(u downto l) := "ZZZZ";
      when 'x' | 'X' => slv(u downto l) := "XXXX";
      when '-'       => slv(u downto i) := "----";
      when '_'       => j := j + 4;   -- filter underscores
      when ' '       => j := j + 4;   -- filter spaces
      when others    => slv(u downto l) := "UUUU";
    end case;
    k := k + 4;
  end loop;
  return slv;
end function hstr_to_slv;


function char_to_sl (c: in character)
  return std_logic is

  variable sl: std_logic;
begin
  case c is
    when 'U' | 'u' => sl := 'U';
    when 'X' | 'x' => sl := 'X';
    when '0'       => sl := '0';
    when '1'       => sl := '1';
    when 'Z' | 'z' => sl := 'Z';
    when 'W' | 'w' => sl := 'W';
    when 'L' | 'l' => sl := 'L';
    when 'H' | 'h' => sl := 'H';
    when '-'       => sl := '-';
    when others    => sl := 'X';
  end case;
  return sl;
end function char_to_sl;


function bstr_to_slv (s: in string)
  return std_logic_vector is

  variable slv: std_logic_vector((s'high-s'low) downto 0);
  variable k: integer;  -- position in std_logic_vector
  variable j: integer;  -- to filter underscores
begin
  k := 0;
  j := 0;
  for i in s'high downto s'low loop
    if (s(i) = '_') or (s(i) = ' ') then  -- filter underscores and spaces
      j := j + 1;
    else
      slv(k-j) := char_to_sl(s(i));
    end if;
    k := k + 1;
  end loop;
  return slv;
end function bstr_to_slv;

procedure proc_str_to_int (
    variable str        : in  string;
    variable int        : out integer;
    variable result     : out boolean
    ) is
   variable v_str     : string(1 to WORD_SIZE);     -- holds clean string value
   variable char_cnt  : integer;  -- number of characters
   variable temp     : integer;
   variable tempint  : integer;

begin
  v_str          := (others => ' ');
  result := true;
  char_cnt := 0;

  for i in str'low to str'high loop
    if (str(i) /= ' ') then  -- skip spaces
      char_cnt          := char_cnt + 1;
      v_str(char_cnt)   := str(i);
    end if;
  end loop;

  tempint := (0);
  for i in 1 to char_cnt loop
    case v_str(i) is
      when '0' => temp := 0;
      when '1' => temp := 1;
      when '2' => temp := 2;
      when '3' => temp := 3;
      when '4' => temp := 4;
      when '5' => temp := 5;
      when '6' => temp := 6;
      when '7' => temp := 7;
      when '8' => temp := 8;
      when '9' => temp := 9;
      when others => result := false;
    end case;
    tempint := tempint + temp * 10**(char_cnt-i);
  end loop;
  int := tempint;

end proc_str_to_int;

function func_lowercase (word : in word_subtype)
  return word_subtype is

  variable temp_word   : word_subtype;
  variable temp_int    : integer;
begin
  temp_word := (others => ' ');
  for i in word'range loop
    temp_int := character'pos(word(i));
    case temp_int is
      when 65 to 90 => temp_word(i) := character'val(temp_int + 32);
      when others => temp_word(i) := word(i);
    end case;
  end loop;
  return temp_word;
end function func_lowercase;


procedure proc_split_line (
  variable line_buf       : in  line;
  variable word_array     : out word_type;
  variable size_array     : out size_type;
  variable word_cnt       : out integer;
  variable empty_line     : out boolean;
  variable comment_line   : out boolean
) is

  variable line_char_cnt  : integer;    -- counts characters in line
  variable word_char_cnt  : integer;    -- counts characters in word
  variable line_word_cnt  : integer;    -- counts words in line
  variable word           : word_subtype;  -- string
  variable stop           : boolean;
begin
  comment_line   := false;
  line_word_cnt  := 0;                    -- clear word counter for new line
  line_char_cnt  := 0;                    -- start at first character in line
  stop           := false;

  -- clear all words in array
  for i in word_array'range loop
    word_array(i) := (others => ' ');
  end loop;

  -- loop while not end of line and number of words is less than max words
  -- in array and it is not an empty line
  while ((line_char_cnt < line_buf'length) and not(stop)) loop
    word_char_cnt:= 0;        -- reset for new word
    word := (others => ' ');  -- clear word

    -- read a character while not end of line
    while (line_char_cnt < line_buf'length) loop
      line_char_cnt       := line_char_cnt + 1;  -- go to next character

      -- if character(s) is(are) '*' or "--" or "//" then rest is comment, stop
      if ( (line_buf(line_char_cnt) = '*') or
           ((line_buf(line_char_cnt) = '-') and (line_buf(line_char_cnt+1) = '-')) or
           ((line_buf(line_char_cnt) = '/') and (line_buf(line_char_cnt+1) = '/')) ) then
        -- if no words and no characters are found then comment line
        comment_line := ((line_word_cnt = 0) and (word_char_cnt = 0));
        stop := true;   -- stop reading line
        exit;           -- rest of line is comment exit loop
      end if;

      -- check on word overflow
      if (word_char_cnt = WORD_SIZE) then
        stop := true;   -- stop reading line
        report "ERROR: Word exceeds maximum of " & integer'image(WORD_SIZE) & " characters";
        exit;
      -- check on space or tab or word size then finish word
      elsif ((line_buf(line_char_cnt) = ' ') or (line_buf(line_char_cnt) = ht)) then
        exit;  -- ready with word exit loop

      else  -- found new character for word
        word_char_cnt       := word_char_cnt + 1;
        word(word_char_cnt) := line_buf(line_char_cnt);
      end if;
    end loop;

    -- check on comment
    if (word_char_cnt /= 0) then
      word_array(line_word_cnt) := func_lowercase(word);
      size_array(line_word_cnt) := word_char_cnt;
      line_word_cnt             := line_word_cnt + 1;

      -- print words found
      --if DEBUG then
      --  report "global_pkg - proc_split_line -> Word " & integer'image(line_word_cnt) & ": " & func_lowercase(word);
      --end if;
    end if;

  end loop;

  empty_line  := (line_word_cnt = 0);  -- set empty line to true when no words found, else false
  word_cnt    := line_word_cnt;        -- retrun number of words

end procedure proc_split_line;

procedure proc_get_value (
    variable str            : in  string;
    variable slv            : out std_logic_vector;
    variable result         : out boolean;
    variable len            : out integer
) is

  variable v_str          : string(1 to WORD_SIZE);     -- clean string value
  variable v_len          : integer;    -- length of slv
  variable v_result       : boolean;    -- needed for result of string to integer conversion
  variable v_found_quote  : boolean;    -- indcates first quote found
  variable v_char_cnt     : integer;    -- counts characters in clean string
begin
  v_str          := (others => ' ');
  v_char_cnt     := 0;
  v_len          := 0;
  v_found_quote  := false;

  -- create a clean string without spaces and underscores, find number of characters
  for i in str'low to str'high loop
    if ((str(i) /= ' ') and (str(i) /= '_') and (str(i) /= CR)) then  -- skip space and underscores
      if (str(i) = '"') then -- " -- check for first quotes in string
        if not(v_found_quote) then  -- skip last quote
          v_found_quote     := true;
          proc_str_to_int(str(str'low to i-2), v_len, v_result);  -- get bit length
          v_str(1)          := str(i-1);  -- first char indicates format
          v_char_cnt        := 2;  -- continue with charater 2 in result string
        end if;
      else
        v_char_cnt        := v_char_cnt + 1;
        v_str(v_char_cnt) := str(i);
      end if;
    end if;
  end loop;

  -- set std logic vector output to undefined
  for i in slv'range loop
    slv(i) := '0';
  end loop;
  len := 0;

  -- determine format
  if (v_found_quote) then
    if ( (v_str(1) = 'x') or (v_str(1) = 'X') ) then  -- found #x"<hex value>"
      slv(v_char_cnt*4-5 downto 0) := hstr_to_slv(v_str(2 to v_char_cnt));  -- remove x
      len := (v_char_cnt-1)*4;
    elsif ( (v_str(1) = 'b') or (v_str(1) = 'B') ) then  -- found #b"<hex value>"
      slv(v_char_cnt-2 downto 0) := bstr_to_slv(v_str(2 to v_char_cnt));  -- remove b"
      len := v_char_cnt-1;
    elsif ( (v_str(1) = 'a') or (v_str(1) = 'A') ) then  -- found #a"<ascii value>"
      slv(v_char_cnt-2 downto 0) := astr_to_slv(v_str(2 to v_char_cnt));  -- remove a"
      len := (v_char_cnt-1)*8;
    end if;
  else
    if ( (v_str(1) = '0') and ((v_str(2) = 'x') or (v_str(2) = 'X')) ) then  -- found 0x<hex value>
      slv(v_char_cnt*4-9 downto 0) := hstr_to_slv(v_str(3 to v_char_cnt));  -- remove 0x
      len := (v_char_cnt-2)*4;
    elsif( (v_str(1) = '0') and ((v_str(2) = 'b') or (v_str(2) = 'B')) ) then  -- found 0b<bin value>
      slv(v_char_cnt-3 downto 0) := bstr_to_slv(v_str(3 to v_char_cnt));  -- remove 0b
      len := v_char_cnt-2;
    else
      slv(v_char_cnt-1 downto 0) := bstr_to_slv(v_str(1 to v_char_cnt));
      len := v_char_cnt;
    end if;
  end if;

  result  := not(v_char_cnt = 0);  -- correct result unless empty string

  -- print value found
  --if DEBUG then
  --  report "global_pkg - proc_get_value -> String value found: " & v_str(1 to v_char_cnt);
  --end if;
end procedure proc_get_value;


function to_hstring (value     : std_logic_vector) return string is
  constant ne     : integer := (value'length+3)/4;
  variable pad    : std_logic_vector(0 to (ne*4 - value'length) - 1);
  variable ivalue : std_logic_vector(0 to ne*4 - 1);
  variable result : string(1 to ne);
  variable quad   : std_logic_vector(0 to 3);
begin
  if value'length < 1 then
    --return NUS;
    return "  ";
  else
    if value (value'left) = 'Z' then
      pad := (others => 'Z');
    else
      pad := (others => '0');
    end if;
    ivalue := pad & value;
    for i in 0 to ne-1 loop
      quad := To_X01Z(ivalue(4*i to 4*i+3));
      case quad is
        when x"0"   => result(i+1) := '0';
        when x"1"   => result(i+1) := '1';
        when x"2"   => result(i+1) := '2';
        when x"3"   => result(i+1) := '3';
        when x"4"   => result(i+1) := '4';
        when x"5"   => result(i+1) := '5';
        when x"6"   => result(i+1) := '6';
        when x"7"   => result(i+1) := '7';
        when x"8"   => result(i+1) := '8';
        when x"9"   => result(i+1) := '9';
        when x"A"   => result(i+1) := 'A';
        when x"B"   => result(i+1) := 'B';
        when x"C"   => result(i+1) := 'C';
        when x"D"   => result(i+1) := 'D';
        when x"E"   => result(i+1) := 'E';
        when x"F"   => result(i+1) := 'F';
        when "ZZZZ" => result(i+1) := 'Z';
        when others => result(i+1) := 'X';
      end case;
    end loop;
    return result;
  end if;
end function to_hstring;




function to_intstring (value     : std_logic_vector) return string is
  --constant ne     : integer := (value'length+3)/4;
  constant ne     : integer := (value'length);

  --variable pad    : std_logic_vector(0 to (ne*4 - value'length) - 1);
  --variable ivalue : std_logic_vector(0 to ne*4 - 1);
  variable result : string(1 to ne);
  --variable quad   : std_logic_vector(0 to 3);

  variable v_int_result : integer;

begin
  if value'length < 1 then
    --return NUS;
    return "  ";
  else
    for i in 0 to ne-1 loop
      if (i=0) then
        if (value(i) = '1') then
          v_int_result := v_int_result + 1;
        else
          v_int_result := v_int_result;
        end if;
      else
        if (value(i) = '1') then
          v_int_result := v_int_result + (i*2);
        end if;
      end if;
    end loop;

    result := integer'image(v_int_result);

  return result;
  end if;
end function to_intstring;


function int_to_string( input : integer ) return string is
  variable my_string : line;
  variable length : integer;
begin
    length := integer( input / 10 + 1 );
    my_string := new string( 1 to length );
    write( my_string, input );
    return my_string.all;
end int_to_string;

procedure str_read(file in_file: TEXT;
                   res_string: out string) is

       variable l:         line;
       variable c:         character;
       variable is_string: boolean;

   begin

     readline(in_file, l);
     -- clear the contents of the result string
     for i in res_string'range loop
         res_string(i) := ' ';
     end loop;
     -- read all characters of the line, up to the length
     -- of the results string
     for i in res_string'range loop
        read(l, c, is_string);
        res_string(i) := c;
        if not is_string then -- found end of line
           exit;
        end if;
     end loop;

end str_read;


end package body tb_env_pkg;
