-- ----------------------------------------------------------------------------
-- Copyright Message
-- ----------------------------------------------------------------------------
--
-- Topic Embedded Systems confidential and proprietary.
-- COPYRIGHT (c) 2009 Topic Embedded Systems B.V.
--
-- All rights are reserved. Reproduction or dissemination
-- in whole or in part is prohibited without the prior written
-- consent of the copyright holder.
--
-- ----------------------------------------------------------------------------
-- Design Information
-- ----------------------------------------------------------------------------
--
-- File : tb_scheduler.vhd
--
-- Author : $Author$
--
-- Organisation : Topic Embedded Systems
--
-- Date : $Date$
--
-- Revision : $Revision$
--
-- Project : Dyplo
--
-- Description :  This unit is part a VHDL test environment and performs
--                register read and writes. Read are checked
--                on expected result and errors are logged.
-- ----------------------------------------------------------------------------
-- Design Notes
-- ----------------------------------------------------------------------------
--
-- <Important notes about design>
-- <signals are active-low or active-high>
-- <main clock signal>
-- <main reset signal>
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

use std.env.all;
use std.textio.all;

use work.tb_env_pkg.all;

library dyplo;
use dyplo.dyplo_types.all;
use dyplo.dyplo_config.all;
use dyplo.dyplo_components.all;

library bfm;
use bfm.bfm_axi4_full_master.all;

entity tb_scheduler is
   port(
      clk                         : in  std_logic;
      rst_n                       : in  std_logic;
      -- control node
      control_cmd_valid           : in  std_logic;
      control_cmd                 : in  std_logic_vector(99 downto 0);
      control_cmd_ready           : out std_logic;
      control_waiting_on_sync     : in  std_logic;
      control_continue_after_sync : out std_logic;
      control_eof                 : in  std_logic;
      -- cpu node
      cpu_cmd_valid               : in  std_logic;
      cpu_cmd                     : in  std_logic_vector(99 downto 0);
      cpu_cmd_ready               : out std_logic;
      cpu_waiting_on_sync         : in  std_logic;
      cpu_continue_after_sync     : out std_logic;
      cpu_eof                     : in  std_logic;
      -- fifo cmd's
      fifo_cmd_valid              : in  std_logic_vector(NO_FIFOS - 1 downto 0);
      fifo_cmd                    : in  fifo_cmd_type;
      fifo_cmd_ready              : out std_logic_vector(NO_FIFOS - 1 downto 0);
      fifo_waiting_on_sync        : in  std_logic_vector(NO_FIFOS - 1 downto 0);
      fifo_continue_after_sync    : out std_logic_vector(NO_FIFOS - 1 downto 0);
      fifo_eof                    : in  std_logic_vector(NO_FIFOS - 1 downto 0);
      -- hdl nodes
      hdl_cmd_valid               : in  std_logic;
      hdl_cmd                     : in  std_logic_vector(99 downto 0);
      hdl_cmd_ready               : out std_logic;
      hdl_waiting_on_sync         : in  std_logic;
      hdl_continue_after_sync     : out std_logic;
      hdl_eof                     : in  std_logic;
      -- dyplo irq
      dyplo_irq                   : in  std_logic;
      -- Interface to AXI VIP commander
      wr_cmd_valid                : out std_logic;
      wr_cmd_address              : out std_logic_vector(31 downto 0);
      wr_cmd_length               : out std_logic_vector(7 downto 0);
      wr_cmd_data                 : out std_logic_vector(32767 downto 0);
      wr_cmd_done                 : in  std_logic;
      rd_cmd_valid                : out std_logic;
      rd_cmd_address              : out std_logic_vector(31 downto 0);
      rd_cmd_length               : out std_logic_vector(7 downto 0);
      rd_cmd_done                 : in  std_logic;
      rd_cmd_data                 : in  std_logic_vector(32767 downto 0)
   );
end tb_scheduler;


architecture behaviour of tb_scheduler is

   constant c_base_addr : std_logic_vector(31 downto 0) := X"64400000";

   constant c_irq_nodes_reg_addr   : std_logic_vector(31 downto 0) := c_base_addr + X"004";
   constant c_irq_status_regs_addr : std_logic_vector(31 downto 0) := c_base_addr + X"824";
   constant c_irq_clear_reg_w_addr : std_logic_vector(31 downto 0) := c_base_addr + X"82C";
   constant c_irq_clear_reg_r_addr : std_logic_vector(31 downto 0) := c_base_addr + X"83C";
   constant c_irq_set_reg_w_addr   : std_logic_vector(31 downto 0) := c_base_addr + X"828";
   constant c_irq_set_reg_r_addr   : std_logic_vector(31 downto 0) := c_base_addr + X"838";

   constant c_no_process    : std_logic_vector(2 downto 0) := "000";
   constant c_process_irq   : std_logic_vector(2 downto 0) := "001";
   constant c_process_read  : std_logic_vector(2 downto 0) := "010";
   constant c_process_write : std_logic_vector(2 downto 0) := "011";
   constant c_process_nodes : std_logic_vector(2 downto 0) := "100";

   constant c_control_node : std_logic_vector(1 downto 0) := "00";
   constant c_cpu_node     : std_logic_vector(1 downto 0) := "01";
   constant c_hdl_node     : std_logic_vector(1 downto 0) := "10";
   constant c_fifos        : std_logic_vector(1 downto 0) := "11";

   constant OKAY   : std_logic_vector(1 downto 0) := "00";                                                    -- AXI RESPONSE OKAY
   constant EXOKAY : std_logic_vector(1 downto 0) := "01";                                                    -- AXI RESPONSE EXOKAY
   constant SLVERR : std_logic_vector(1 downto 0) := "10";                                                    -- AXI RESPONSE SLVERR
   constant DECERR : std_logic_vector(1 downto 0) := "11";                                                    -- AXI RESPONSE DECERR

   constant BFM_CMD_DATA_WIDTH : integer := 100;

   type sm_bfm_write_type is (WAIT_FOR_CMD, WAIT_FOR_BFM_WRITE);
   signal sm_bfm_write : sm_bfm_write_type := WAIT_FOR_CMD;

   type sm_bfm_read_type is (WAIT_FOR_CMD, WAIT_FOR_BFM_READ);
   signal sm_bfm_read : sm_bfm_read_type := WAIT_FOR_CMD;

   type sm_scheduler_type is (IDLE, WAIT_FOR_IRQ_INFO, REQUEST_DATA, WAIT_AND_LOG_DATA, PARSE_CMD, WAIT_FOR_BFM, LOG_DATA, FINISH_IRQ);
   signal sm_scheduler : sm_scheduler_type := IDLE;

   type sm_irq_type is (IDLE, WAIT_FOR_IRQ_INFO, CLEAR_READ_IRQ, CLEAR_WRITE_IRQ, WAIT_FOR_WRITE_IRQ_CLEARED, WAIT_FOR_READ_IRQ_CLEARED);
   signal sm_irq : sm_irq_type := IDLE;

   type sm_fifo_write_handler_type is (IDLE, HANDLE_FIFO_CMD, INITIATE_WRITE_CMD, LOG_DATA, INITIATE_STATUS_READ, WAIT_FOR_STATUS_READ, WAIT_FOR_BFM_WRITE, SET_THRESHOLD, WAIT_FOR_THRESHOLD_SET, WAIT_FOR_IRQ_SET);
   signal sm_fifo_write_handler : sm_fifo_write_handler_type := IDLE;

   type sm_fifo_read_handler_type is (IDLE, HANDLE_FIFO_CMD, INITIATE_READ_CMD, LOG_DATA, INITIATE_STATUS_READ, WAIT_FOR_STATUS_READ, WAIT_FOR_BFM_READ, SET_THRESHOLD, WAIT_FOR_THRESHOLD_SET, WAIT_FOR_IRQ_SET);
   signal sm_fifo_read_handler : sm_fifo_read_handler_type := IDLE;

   type sm_node_handler_type is (IDLE, INITIATE_CMD, WAIT_FOR_BFM_READ, WAIT_FOR_BFM_WRITE, LOG_DATA);
   signal sm_node_handler : sm_node_handler_type := IDLE;

   type sm_syncer_type is (WAITING_FOR_SYNC, WAITING_FOR_CTRL, WAITING_FOR_CPU, WAITING_FOR_HDL);
   signal sm_syncer : sm_syncer_type := WAITING_FOR_SYNC;

   signal bfm_write_done : boolean := false;
   signal bfm_read_done  : boolean := false;

   signal bfm_write_cmd       : std_logic_vector((BFM_CMD_DATA_WIDTH - 1) downto 0) := (others => '0');
   --(3:0) command
   --(35:4) address
   --(67:36) data / expected data / filename (x.txt)
   --(99:68) length in words
   signal bfm_write_cmd_valid : std_logic                                           := '0';
   signal bfm_write_token     : std_logic_vector(2 downto 0);
   signal bfm_write_busy      : std_logic                                           := '0';

   signal bfm_read_cmd       : std_logic_vector((BFM_CMD_DATA_WIDTH - 1) downto 0) := (others => '0');
   --(3:0) command
   --(35:4) address
   --(67:36) data / expected data / filename (x.txt)
   --(99:68) length in words
   signal bfm_read_cmd_valid : std_logic                                           := '0';
   signal bfm_read_token     : std_logic_vector(2 downto 0);
   signal bfm_read_busy      : std_logic                                           := '0';


   signal bfm_write_cmd_i       : std_logic_vector((BFM_CMD_DATA_WIDTH - 1) downto 0) := (others => '0');
   signal bfm_write_cmd_valid_i : std_logic                                           := '0';
   signal bfm_read_cmd_i        : std_logic_vector((BFM_CMD_DATA_WIDTH - 1) downto 0) := (others => '0');
   signal bfm_read_cmd_valid_i  : std_logic                                           := '0';

   signal bfm_write_cmd_w       : std_logic_vector((BFM_CMD_DATA_WIDTH - 1) downto 0) := (others => '0');
   signal bfm_write_cmd_valid_w : std_logic                                           := '0';
   signal bfm_read_cmd_w        : std_logic_vector((BFM_CMD_DATA_WIDTH - 1) downto 0) := (others => '0');
   signal bfm_read_cmd_valid_w  : std_logic                                           := '0';

   signal bfm_write_cmd_r       : std_logic_vector((BFM_CMD_DATA_WIDTH - 1) downto 0) := (others => '0');
   signal bfm_write_cmd_valid_r : std_logic                                           := '0';
   signal bfm_read_cmd_r        : std_logic_vector((BFM_CMD_DATA_WIDTH - 1) downto 0) := (others => '0');
   signal bfm_read_cmd_valid_r  : std_logic                                           := '0';

   signal bfm_write_cmd_nodes       : std_logic_vector((BFM_CMD_DATA_WIDTH - 1) downto 0) := (others => '0');
   signal bfm_write_cmd_valid_nodes : std_logic                                           := '0';
   signal bfm_read_cmd_nodes        : std_logic_vector((BFM_CMD_DATA_WIDTH - 1) downto 0) := (others => '0');
   signal bfm_read_cmd_valid_nodes  : std_logic                                           := '0';

   signal bfm_read_req_w     : std_logic := '0';
   signal bfm_read_req_r     : std_logic := '0';
   signal bfm_write_req_w    : std_logic := '0';
   signal bfm_write_req_r    : std_logic := '0';
   signal switch_write_token : std_logic := '1';
   signal switch_read_token  : std_logic := '1';

   signal arbitter   : std_logic_vector(1 downto 0) := "00";
   signal sync_delay : std_logic                    := '1';

   signal handle_irq : std_logic := '0';

   signal i_w : integer := 0;
   signal i_r : integer := 0;

   type addr_write_type is array (NO_FIFOS - 1 downto 0) of std_logic_vector(31 downto 0);
   signal addr_write       : addr_write_type;
   type data_write_type is array (NO_FIFOS - 1 downto 0) of std_logic_vector(31 downto 0);
   signal data_write       : data_write_type;
   type length_write_type is array (NO_FIFOS - 1 downto 0) of std_logic_vector(31 downto 0);
   signal length_write     : length_write_type;
   type datafileID_write_type is array (NO_FIFOS - 1 downto 0) of std_logic_vector(31 downto 0);
   signal datafileID_write : datafileID_write_type;
   type space_in_fifo_type is array (NO_FIFOS - 1 downto 0) of std_logic_vector(11 downto 0);
   signal space_in_fifo    : space_in_fifo_type;
   signal fifo_cmd_ready_w : std_logic_vector(NO_FIFOS - 1 downto 0) := (others => '0');

   type addr_read_type is array (NO_FIFOS - 1 downto 0) of std_logic_vector(31 downto 0);
   signal addr_read        : addr_read_type;
   type expected_data_type is array (NO_FIFOS - 1 downto 0) of std_logic_vector(31 downto 0);
   signal expected_data    : expected_data_type;
   type length_read_type is array (NO_FIFOS - 1 downto 0) of std_logic_vector(31 downto 0);
   signal length_read      : length_read_type;
   type datafileID_read_type is array (NO_FIFOS - 1 downto 0) of std_logic_vector(31 downto 0);
   signal datafileID_read  : datafileID_read_type;
   type words_in_fifo_type is array (NO_FIFOS - 1 downto 0) of std_logic_vector(11 downto 0);
   signal words_in_fifo    : words_in_fifo_type;
   signal fifo_cmd_ready_r : std_logic_vector(NO_FIFOS - 1 downto 0) := (others => '0');

   type words_send_type is array (NO_FIFOS - 1 downto 0) of integer;
   signal words_send : words_send_type;

   type handling_cmd_type is array (NO_FIFOS - 1 downto 0) of std_logic;
   signal handling_cmd_w : handling_cmd_type;
   signal handling_cmd_r : handling_cmd_type;

   signal RD_DATA_from_master : std_logic_vector((32 * (255 + 1)) - 1 downto 0) := (others => '0');

begin

   -- open log and stimuli files
   process
   begin
      file_open(v_log_file_status, log_file, RESULTS_LOG_FILE, write_mode);
      if not (v_log_file_status = OPEN_OK) then
         report "ERROR: Unable to open log file " & RESULTS_LOG_FILE
         severity FAILURE;
      end if;
      wait;
   end process;

   p_irq_handler : process(clk)
      variable v_log_line : line;                                                                             -- store log line
   begin
      if (rising_edge(clk)) then
         if (rst_n = '0') then
         else
            case (sm_irq) is
               when IDLE =>

                  if (dyplo_irq = '1') then
                     handle_irq <= '1';
                     if (bfm_read_busy /= '1' and bfm_read_token = c_process_irq) then
                        bfm_read_cmd_i(3 downto 0) <= c_req_irq_info;
                        bfm_read_cmd_valid_i       <= '1';

                        sm_irq <= WAIT_FOR_IRQ_INFO;
                     end if;
                  else
                     handle_irq <= '0';
                  end if;

               when WAIT_FOR_IRQ_INFO =>
                  bfm_read_cmd_valid_i <= '0';
                  if (bfm_read_done = true) then

                     write(v_log_line, string'("Dyplo IRQ triggered with read_mask: " & to_hstring(RD_DATA_from_master(31 downto 0)) & ", write_mask: " & to_hstring(RD_DATA_from_master(159 downto 128))));
                     writeline(log_file, v_log_line);

                     if (RD_DATA_from_master(159 downto 128) /= X"00000000") then
                        sm_irq <= CLEAR_READ_IRQ;
                     elsif (RD_DATA_from_master(31 downto 0) /= X"00000000") then
                        sm_irq <= CLEAR_WRITE_IRQ;
                     else
                        sm_irq <= IDLE;
                     end if;
                  end if;

               when CLEAR_READ_IRQ =>
                  if (bfm_write_busy /= '1' and bfm_write_token = c_process_irq) then
                     bfm_write_cmd_i(3 downto 0)  <= c_clear_read_irq;
                     bfm_write_cmd_i(35 downto 4) <= RD_DATA_from_master(159 downto 128);
                     bfm_write_cmd_valid_i        <= '1';

                     sm_irq <= WAIT_FOR_READ_IRQ_CLEARED;
                  end if;

               when CLEAR_WRITE_IRQ =>
                  if (bfm_write_busy /= '1' and bfm_write_token = c_process_irq) then
                     bfm_write_cmd_i(3 downto 0)  <= c_clear_write_irq;
                     bfm_write_cmd_i(35 downto 4) <= RD_DATA_from_master(31 downto 0);
                     bfm_write_cmd_valid_i        <= '1';

                     handle_irq <= '0';
                     sm_irq     <= WAIT_FOR_WRITE_IRQ_CLEARED;
                  end if;

               when WAIT_FOR_READ_IRQ_CLEARED =>
                  bfm_write_cmd_valid_i <= '0';
                  if (bfm_write_done = true) then
                     if (RD_DATA_from_master(31 downto 0) /= X"00000000") then
                        sm_irq <= CLEAR_WRITE_IRQ;
                     else
                        sm_irq <= IDLE;
                     end if;
                  end if;

               when WAIT_FOR_WRITE_IRQ_CLEARED =>
                  bfm_write_cmd_valid_i <= '0';
                  if (bfm_write_done = true) then
                     sm_irq <= IDLE;
                  end if;
            end case;
         end if;
      end if;
   end process p_irq_handler;


   p_nodes_write_handler : process(clk)
      variable v_log_line : line;                                                                             -- store log line
   begin
      if (rising_edge(clk)) then
         if rst_n = '0' then
            control_cmd_ready <= '0';
            cpu_cmd_ready     <= '0';
            hdl_cmd_ready     <= '0';
         else
            if (arbitter = c_control_node) then

               case (sm_node_handler) is

                  when IDLE =>
                     control_cmd_ready <= '0';
                     sm_node_handler   <= INITIATE_CMD;

                  when INITIATE_CMD =>

                     if (control_cmd_valid = '1') then
                        if (control_cmd(3 downto 0) = c_write_conf) then
                           if (bfm_write_busy /= '1' and bfm_write_token = c_process_nodes) then
                              bfm_write_cmd_nodes(3 downto 0)   <= c_write_conf;
                              bfm_write_cmd_nodes(35 downto 4)  <= control_cmd(35 downto 4);
                              bfm_write_cmd_nodes(67 downto 36) <= control_cmd(67 downto 36);
                              bfm_write_cmd_valid_nodes         <= '1';

                              sm_node_handler <= WAIT_FOR_BFM_WRITE;
                           end if;
                        elsif (control_cmd(3 downto 0) = c_read_conf_no_check) then
                           if (bfm_read_busy /= '1' and bfm_read_token = c_process_nodes) then
                              bfm_read_cmd_nodes(3 downto 0)  <= c_read_conf_no_check;
                              bfm_read_cmd_nodes(35 downto 4) <= control_cmd(35 downto 4);
                              bfm_read_cmd_valid_nodes        <= '1';

                              sm_node_handler <= WAIT_FOR_BFM_READ;
                           end if;

                        elsif (control_cmd(3 downto 0) = c_read_conf_check_result) then
                           if (bfm_read_busy /= '1' and bfm_read_token = c_process_nodes) then
                              bfm_read_cmd_nodes(3 downto 0)  <= c_read_conf_no_check;
                              bfm_read_cmd_nodes(35 downto 4) <= control_cmd(35 downto 4);
                              bfm_read_cmd_valid_nodes        <= '1';

                              sm_node_handler <= WAIT_FOR_BFM_READ;
                           end if;
                        end if;
                     end if;

                  when WAIT_FOR_BFM_WRITE =>
                     bfm_write_cmd_valid_nodes <= '0';
                     if (bfm_write_done = true) then
                        sm_node_handler <= LOG_DATA;
                     end if;

                  when WAIT_FOR_BFM_READ =>
                     bfm_read_cmd_valid_nodes <= '0';
                     if (bfm_read_done = true) then
                        sm_node_handler <= LOG_DATA;
                     end if;

                  when LOG_DATA =>

                     if (control_cmd(3 downto 0) = c_write_conf) then

                        write(v_log_line, string'("Control write to address 0x"));
                        write(v_log_line, to_hstring(control_cmd(35 downto 4)));
                        write(v_log_line, string'(": 0x") & to_hstring(control_cmd(67 downto 36)));
                        case (RESPONSE) is
                           when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                           when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                           when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                           when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                           when others => write(v_log_line, string'(", AXI:Undefined Response"));
                        end case;
                        writeline(log_file, v_log_line);

                        control_cmd_ready <= '1';

                     elsif (control_cmd(3 downto 0) = c_read_conf_no_check) then

                        write(v_log_line, string'("Control read from address 0x"));
                        write(v_log_line, to_hstring(control_cmd(35 downto 4)));
                        write(v_log_line, string'(": 0x") & to_hstring(RD_DATA_from_master(31 downto 0)));
                        case (VRESPONSE(1 downto 0)) is
                           when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                           when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                           when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                           when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                           when others => write(v_log_line, string'(", AXI:Undefined Response"));
                        end case;
                        writeline(log_file, v_log_line);

                        control_cmd_ready <= '1';

                     elsif (control_cmd(3 downto 0) = c_read_conf_check_result) then

                        if (control_cmd(67 downto 36) = (RD_DATA_from_master(31 downto 0) and control_cmd(99 downto 68))) then
                           write(v_log_line, string'("Control read from address 0x"));
                           write(v_log_line, to_hstring(control_cmd(35 downto 4)));
                           write(v_log_line, string'(": 0x") & to_hstring(RD_DATA_from_master(31 downto 0)));
                           write(v_log_line, string'(", ok"));
                           report "*** Data read from register OK";
                        else
                           write(v_log_line, string'("ERROR: "));
                           write(v_log_line, string'("Control read from address 0x"));
                           write(v_log_line, to_hstring(control_cmd(35 downto 4)));
                           write(v_log_line, string'(": 0x") & to_hstring(RD_DATA_from_master(31 downto 0)));
                           write(v_log_line, string'(", expected "));
                           write(v_log_line, string'("0x") & to_hstring(control_cmd(67 downto 36)));
                           write(v_log_line, string'(" (mask:  "));
                           write(v_log_line, string'("0x") & to_hstring(control_cmd(99 downto 68)));
                           report "*** Data read from register NOK!";
                        end if;

                        case (VRESPONSE(1 downto 0)) is
                           when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                           when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                           when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                           when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                           when others => write(v_log_line, string'(", AXI:Undefined Response"));
                        end case;
                        writeline(log_file, v_log_line);

                        control_cmd_ready <= '1';
                     end if;

                     sm_node_handler <= IDLE;
               end case;

            elsif (arbitter = c_cpu_node) then

               case (sm_node_handler) is

                  when IDLE =>
                     cpu_cmd_ready   <= '0';
                     sm_node_handler <= INITIATE_CMD;

                  when INITIATE_CMD =>
                     if (cpu_cmd_valid = '1') then
                        if (cpu_cmd(3 downto 0) = c_write_conf) then
                           if (bfm_write_busy /= '1' and bfm_write_token = c_process_nodes) then
                              bfm_write_cmd_nodes(3 downto 0)   <= c_write_conf;
                              bfm_write_cmd_nodes(35 downto 4)  <= cpu_cmd(35 downto 4);
                              bfm_write_cmd_nodes(67 downto 36) <= cpu_cmd(67 downto 36);
                              bfm_write_cmd_valid_nodes         <= '1';

                              sm_node_handler <= WAIT_FOR_BFM_WRITE;
                           end if;
                        elsif (cpu_cmd(3 downto 0) = c_read_conf_no_check) then
                           if (bfm_read_busy /= '1' and bfm_read_token = c_process_nodes) then
                              bfm_read_cmd_nodes(3 downto 0)  <= c_read_conf_no_check;
                              bfm_read_cmd_nodes(35 downto 4) <= cpu_cmd(35 downto 4);
                              bfm_read_cmd_valid_nodes        <= '1';

                              sm_node_handler <= WAIT_FOR_BFM_READ;
                           end if;

                        elsif (cpu_cmd(3 downto 0) = c_read_conf_check_result) then
                           if (bfm_read_busy /= '1' and bfm_read_token = c_process_nodes) then
                              bfm_read_cmd_nodes(3 downto 0)  <= c_read_conf_no_check;
                              bfm_read_cmd_nodes(35 downto 4) <= cpu_cmd(35 downto 4);
                              bfm_read_cmd_valid_nodes        <= '1';

                              sm_node_handler <= WAIT_FOR_BFM_READ;
                           end if;
                        end if;
                     end if;

                  when WAIT_FOR_BFM_WRITE =>
                     bfm_write_cmd_valid_nodes <= '0';
                     if (bfm_write_done = true) then
                        sm_node_handler <= LOG_DATA;
                     end if;

                  when WAIT_FOR_BFM_READ =>
                     bfm_read_cmd_valid_nodes <= '0';
                     if (bfm_read_done = true) then
                        sm_node_handler <= LOG_DATA;
                     end if;

                  when LOG_DATA =>

                     if (cpu_cmd(3 downto 0) = c_write_conf) then

                        write(v_log_line, string'("CPU write to address 0x"));
                        write(v_log_line, to_hstring(cpu_cmd(35 downto 4)));
                        write(v_log_line, string'(": 0x") & to_hstring(cpu_cmd(67 downto 36)));
                        case (RESPONSE) is
                           when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                           when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                           when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                           when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                           when others => write(v_log_line, string'(", AXI:Undefined Response"));
                        end case;
                        writeline(log_file, v_log_line);

                        cpu_cmd_ready <= '1';

                     elsif (cpu_cmd(3 downto 0) = c_read_conf_no_check) then

                        write(v_log_line, string'("CPU read from address 0x"));
                        write(v_log_line, to_hstring(cpu_cmd(35 downto 4)));
                        write(v_log_line, string'(": 0x") & to_hstring(RD_DATA_from_master(31 downto 0)));
                        case (VRESPONSE(1 downto 0)) is
                           when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                           when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                           when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                           when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                           when others => write(v_log_line, string'(", AXI:Undefined Response"));
                        end case;
                        writeline(log_file, v_log_line);

                        cpu_cmd_ready <= '1';

                     elsif (cpu_cmd(3 downto 0) = c_read_conf_check_result) then

                        if (cpu_cmd(67 downto 36) = (RD_DATA_from_master(31 downto 0) and cpu_cmd(99 downto 68))) then
                           write(v_log_line, string'("CPU read from address 0x"));
                           write(v_log_line, to_hstring(cpu_cmd(35 downto 4)));
                           write(v_log_line, string'(": 0x") & to_hstring(RD_DATA_from_master(31 downto 0)));
                           write(v_log_line, string'(", ok"));
                           report "*** Data read from register OK";
                        else
                           write(v_log_line, string'("ERROR: "));
                           write(v_log_line, string'("CPU read from address 0x"));
                           write(v_log_line, to_hstring(cpu_cmd(35 downto 4)));
                           write(v_log_line, string'(": 0x") & to_hstring(RD_DATA_from_master(31 downto 0)));
                           write(v_log_line, string'(", expected "));
                           write(v_log_line, string'("0x") & to_hstring(cpu_cmd(67 downto 36)));
                           write(v_log_line, string'(" (mask:  "));
                           write(v_log_line, string'("0x") & to_hstring(cpu_cmd(99 downto 68)));
                           write(v_log_line, string'(")."));
                           report "*** Data read from register NOK!";
                        end if;

                        case (VRESPONSE(1 downto 0)) is
                           when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                           when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                           when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                           when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                           when others => write(v_log_line, string'(", AXI:Undefined Response"));
                        end case;
                        writeline(log_file, v_log_line);

                        cpu_cmd_ready <= '1';
                     end if;

                     sm_node_handler <= IDLE;
               end case;

            elsif (arbitter = c_hdl_node) then

               case (sm_node_handler) is

                  when IDLE =>
                     hdl_cmd_ready   <= '0';
                     sm_node_handler <= INITIATE_CMD;

                  when INITIATE_CMD =>
                     if (hdl_cmd_valid = '1') then
                        if (hdl_cmd(3 downto 0) = c_write_conf) then
                           if (bfm_write_busy /= '1' and bfm_write_token = c_process_nodes) then
                              bfm_write_cmd_nodes(3 downto 0)   <= c_write_conf;
                              bfm_write_cmd_nodes(35 downto 4)  <= hdl_cmd(35 downto 4);
                              bfm_write_cmd_nodes(67 downto 36) <= hdl_cmd(67 downto 36);
                              bfm_write_cmd_valid_nodes         <= '1';

                              sm_node_handler <= WAIT_FOR_BFM_WRITE;
                           end if;

                        elsif (hdl_cmd(3 downto 0) = c_read_conf_no_check) then
                           if (bfm_read_busy /= '1' and bfm_read_token = c_process_nodes) then
                              bfm_read_cmd_nodes(3 downto 0)  <= c_read_conf_no_check;
                              bfm_read_cmd_nodes(35 downto 4) <= hdl_cmd(35 downto 4);
                              bfm_read_cmd_valid_nodes        <= '1';

                              sm_node_handler <= WAIT_FOR_BFM_READ;
                           end if;

                        elsif (hdl_cmd(3 downto 0) = c_read_conf_check_result) then
                           if (bfm_read_busy /= '1' and bfm_read_token = c_process_nodes) then
                              bfm_read_cmd_nodes(3 downto 0)  <= c_read_conf_no_check;
                              bfm_read_cmd_nodes(35 downto 4) <= hdl_cmd(35 downto 4);
                              bfm_read_cmd_valid_nodes        <= '1';

                              sm_node_handler <= WAIT_FOR_BFM_READ;
                           end if;
                        end if;
                     end if;

                  when WAIT_FOR_BFM_WRITE =>
                     bfm_write_cmd_valid_nodes <= '0';
                     if (bfm_write_done = true) then
                        sm_node_handler <= LOG_DATA;
                     end if;

                  when WAIT_FOR_BFM_READ =>
                     bfm_read_cmd_valid_nodes <= '0';
                     if (bfm_read_done = true) then
                        sm_node_handler <= LOG_DATA;
                     end if;

                  when LOG_DATA =>

                     if (hdl_cmd(3 downto 0) = c_write_conf) then

                        write(v_log_line, string'("HDL write to address 0x"));
                        write(v_log_line, to_hstring(hdl_cmd(35 downto 4)));
                        write(v_log_line, string'(": 0x") & to_hstring(hdl_cmd(67 downto 36)));
                        case (RESPONSE) is
                           when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                           when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                           when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                           when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                           when others => write(v_log_line, string'(", AXI:Undefined Response"));
                        end case;
                        writeline(log_file, v_log_line);

                        hdl_cmd_ready <= '1';

                     elsif (hdl_cmd(3 downto 0) = c_read_conf_no_check) then

                        write(v_log_line, string'("HDL read from address 0x"));
                        write(v_log_line, to_hstring(hdl_cmd(35 downto 4)));
                        write(v_log_line, string'(": 0x") & to_hstring(RD_DATA_from_master(31 downto 0)));
                        case (VRESPONSE(1 downto 0)) is
                           when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                           when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                           when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                           when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                           when others => write(v_log_line, string'(", AXI:Undefined Response"));
                        end case;
                        writeline(log_file, v_log_line);

                        hdl_cmd_ready <= '1';

                     elsif (hdl_cmd(3 downto 0) = c_read_conf_check_result) then

                        if (hdl_cmd(67 downto 36) = (RD_DATA_from_master(31 downto 0) and hdl_cmd(99 downto 68))) then
                           write(v_log_line, string'("HDL read from address 0x"));
                           write(v_log_line, to_hstring(hdl_cmd(35 downto 4)));
                           write(v_log_line, string'(": 0x") & to_hstring(RD_DATA_from_master(31 downto 0)));
                           write(v_log_line, string'(", ok"));
                           report "*** Data read from register OK";
                        else
                           write(v_log_line, string'("ERROR: "));
                           write(v_log_line, string'("HDL read from address 0x"));
                           write(v_log_line, to_hstring(hdl_cmd(35 downto 4)));
                           write(v_log_line, string'(": 0x") & to_hstring(RD_DATA_from_master(31 downto 0)));
                           write(v_log_line, string'(", expected "));
                           write(v_log_line, string'("0x") & to_hstring(hdl_cmd(67 downto 36)));
                           write(v_log_line, string'(" (mask:  "));
                           write(v_log_line, string'("0x") & to_hstring(hdl_cmd(99 downto 68)));
                           write(v_log_line, string'(")."));
                           report "*** Data read from register NOK!";
                        end if;

                        case (VRESPONSE(1 downto 0)) is
                           when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                           when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                           when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                           when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                           when others => write(v_log_line, string'(", AXI:Undefined Response"));
                        end case;
                        writeline(log_file, v_log_line);

                        hdl_cmd_ready <= '1';
                     end if;

                     sm_node_handler <= IDLE;

               end case;
            end if;
         end if;
      end if;
   end process p_nodes_write_handler;

   p_fifo_write_handler : process(clk)
      variable v_log_line : line;                                                                             -- store log line
   begin
      if (rising_edge(clk)) then
         if rst_n = '0' then
            datafileID_write <= (others => (others => '0'));
            data_write       <= (others => (others => '0'));
            length_write     <= (others => (others => '0'));
            space_in_fifo    <= (others => (others => '0'));
            fifo_cmd_ready_w <= (others => '0');
            handling_cmd_w   <= (others => '0');
            words_send       <= (others => 0);
         else

            case (sm_fifo_write_handler) is
               when IDLE =>
                  fifo_cmd_ready_w(i_w) <= '0';
                  if (i_w < NO_FIFOS - 1) then
                     i_w <= i_w + 1;
                  else
                     i_w <= 0;
                  end if;
                  sm_fifo_write_handler <= HANDLE_FIFO_CMD;

               when HANDLE_FIFO_CMD =>

                  if (fifo_eof(i_w) /= '1' and fifo_waiting_on_sync(i_w) /= '1') then
                     if (fifo_cmd_valid(i_w) = '1' and fifo_cmd(i_w)(3) = '1') then
                        if (handling_cmd_w(i_w) = '0') then
                           addr_write(i_w) <= fifo_cmd(i_w)(35 downto 4);

                           if (fifo_cmd(i_w)(3 downto 0) = c_write_conf) then
                              data_write(i_w)       <= fifo_cmd(i_w)(67 downto 36);
                              bfm_write_req_w       <= '1';
                              sm_fifo_write_handler <= INITIATE_WRITE_CMD;

                           elsif (fifo_cmd(i_w)(3 downto 0) = c_write_data) then
                              bfm_read_req_w        <= '1';
                              datafileID_write(i_w) <= fifo_cmd(i_w)(67 downto 36);
                              length_write(i_w)     <= fifo_cmd(i_w)(99 downto 68);

                              handling_cmd_w(i_w) <= '1';

                              bfm_read_req_w        <= '1';
                              sm_fifo_write_handler <= INITIATE_STATUS_READ;
                           else
                              sm_fifo_write_handler <= IDlE;
                           end if;
                        else
                           bfm_read_req_w        <= '1';
                           sm_fifo_write_handler <= INITIATE_STATUS_READ;
                        end if;
                     else
                        sm_fifo_write_handler <= IDLE;
                     end if;
                  else
                     sm_fifo_write_handler <= IDLE;
                  end if;

               when INITIATE_WRITE_CMD =>

                  if (fifo_cmd(i_w)(3 downto 0) = c_write_conf) then
                     if (bfm_write_busy /= '1' and bfm_write_token = c_process_write) then
                        bfm_write_cmd_w(3 downto 0)   <= c_write_conf;
                        bfm_write_cmd_w(35 downto 4)  <= addr_write(i_w);
                        bfm_write_cmd_w(67 downto 36) <= data_write(i_w);
                        bfm_write_cmd_valid_w         <= '1';

                        sm_fifo_write_handler <= WAIT_FOR_BFM_WRITE;
                     end if;

                  elsif (fifo_cmd(i_w)(3 downto 0) = c_write_data) then
                     if (bfm_write_busy /= '1' and bfm_write_token = c_process_write) then
                        if (length_write(i_w) /= 0) then

                           bfm_write_cmd_w(3 downto 0)   <= c_write_data;
                           bfm_write_cmd_w(35 downto 4)  <= addr_write(i_w);
                           bfm_write_cmd_w(67 downto 36) <= datafileID_write(i_w);
                           if (space_in_fifo(i_w) >= length_write(i_w)) then
                              if (length_write(i_w) >= X"00000100") then
                                 bfm_write_cmd_w(99 downto 68) <= X"00000100";
                              else
                                 bfm_write_cmd_w(99 downto 68) <= length_write(i_w);
                              end if;
                           else
                              bfm_write_cmd_w(99 downto 68) <= X"00000" & space_in_fifo(i_w);
                           end if;
                           bfm_write_cmd_valid_w         <= '1';

                           sm_fifo_write_handler <= WAIT_FOR_BFM_WRITE;
                        else
                           fifo_cmd_ready_w(i_w) <= '1';
                           bfm_write_req_w       <= '0';
                           sm_fifo_write_handler <= IDLE;
                        end if;
                     end if;
                  end if;

               when WAIT_FOR_BFM_WRITE =>
                  bfm_write_cmd_valid_w <= '0';
                  if (bfm_write_done = true) then
                     length_write(i_w) <= length_write(i_w) - bfm_write_cmd_w(99 downto 68);
                     words_send(i_w)   <= words_send(i_w) + conv_integer(bfm_write_cmd_w(99 downto 68));

                     bfm_write_req_w       <= '0';
                     sm_fifo_write_handler <= LOG_DATA;
                  end if;

               when LOG_DATA =>

                  if (fifo_cmd(i_w)(3 downto 0) = c_write_conf) then

                     write(v_log_line, string'("Fifo") & integer'image(i_w) & string'(" Write to address 0x"));
                     write(v_log_line, to_hstring(bfm_write_cmd_w(35 downto 4)));
                     write(v_log_line, string'(": 0x") & to_hstring(bfm_write_cmd_w(67 downto 36)));
                     case (RESPONSE) is
                        when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                        when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                        when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                        when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                        when others => write(v_log_line, string'(", AXI:Undefined Response"));
                     end case;
                     writeline(log_file, v_log_line);

                     fifo_cmd_ready_w(i_w) <= '1';
                     sm_fifo_write_handler <= IDLE;

                  elsif (fifo_cmd(i_w)(3 downto 0) = c_write_data) then

                     write(v_log_line, string'("Fifo") & integer'image(i_w) & string'(" Write burst to address 0x"));
                     write(v_log_line, to_hstring(bfm_write_cmd_w(35 downto 4)));
                     write(v_log_line, string'(": ID ") & integer'image(conv_integer(bfm_write_cmd_w(67 downto 36))));
                     write(v_log_line, string'(": length: 0x") & to_hstring(bfm_write_cmd_w(99 downto 68)) & string'(" words"));
                     case (RESPONSE) is
                        when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                        when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                        when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                        when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                        when others => write(v_log_line, string'(", AXI:Undefined Response"));
                     end case;
                     writeline(log_file, v_log_line);

                     --              if(length_write(i_w) > 0) then
                     --                bfm_read_req_w          <= '1';
                     --                sm_fifo_write_handler   <= INITIATE_STATUS_READ;
                     --              else
                     --                handling_cmd_w(i_w)     <= '0';
                     --                fifo_cmd_ready_w(i_w)   <= '1';
                     --                sm_fifo_write_handler   <= IDLE;
                     --              end if;

                     if (length_write(i_w) > 0) then
                     --bfm_read_req_w          <= '1';
                     --sm_fifo_write_handler   <= INITIATE_STATUS_READ;
                     else
                        handling_cmd_w(i_w)   <= '0';
                        fifo_cmd_ready_w(i_w) <= '1';
                        --sm_fifo_write_handler   <= IDLE;
                     end if;

                     sm_fifo_write_handler <= IDLE;

                  end if;

               when INITIATE_STATUS_READ =>
                  if (bfm_read_busy /= '1' and bfm_read_token = c_process_write) then
                     bfm_read_cmd_w(3 downto 0)  <= c_req_write_fifo_status;
                     bfm_read_cmd_w(35 downto 4) <= std_logic_vector(to_unsigned(i_w, 32));
                     bfm_read_cmd_valid_w        <= '1';

                     sm_fifo_write_handler <= WAIT_FOR_STATUS_READ;
                  end if;

               when WAIT_FOR_STATUS_READ =>
                  bfm_read_cmd_valid_w <= '0';
                  if (bfm_read_done = true) then
                     if (RD_DATA_from_master(11 downto 0) = X"000") then
                        bfm_write_req_w       <= '1';
                        sm_fifo_write_handler <= SET_THRESHOLD;
                     else
                        space_in_fifo(i_w)    <= RD_DATA_from_master(11 downto 0);
                        bfm_write_req_w       <= '1';
                        sm_fifo_write_handler <= INITIATE_WRITE_CMD;
                     end if;

                     bfm_read_req_w <= '0';

                  end if;

               when SET_THRESHOLD =>
                  if (bfm_write_busy /= '1' and bfm_write_token = c_process_write) then
                     bfm_write_cmd_w(3 downto 0)  <= c_set_write_threshold;
                     bfm_write_cmd_w(35 downto 4) <= std_logic_vector(to_unsigned(i_w, 32));
                     if (length_write(i_w) >= (c_cpu_write_fifo_size / 2)) then
                        bfm_write_cmd_w(67 downto 36) <= std_logic_vector(to_unsigned(c_cpu_write_fifo_size / 2, 32));
                     else
                        bfm_write_cmd_w(67 downto 36) <= length_write(i_w);
                     end if;
                     bfm_write_cmd_valid_w        <= '1';

                     sm_fifo_write_handler <= WAIT_FOR_THRESHOLD_SET;
                  end if;

               when WAIT_FOR_THRESHOLD_SET =>
                  bfm_write_cmd_valid_w <= '0';
                  if (bfm_write_done = true) then
                     bfm_write_cmd_w(3 downto 0)  <= c_set_write_irq;
                     bfm_write_cmd_w(35 downto 4) <= std_logic_vector(to_unsigned(i_w, 32));
                     bfm_write_cmd_valid_w        <= '1';

                     sm_fifo_write_handler <= WAIT_FOR_IRQ_SET;
                  end if;

               when WAIT_FOR_IRQ_SET =>
                  bfm_write_cmd_valid_w <= '0';
                  if (bfm_write_done = true) then
                     bfm_write_req_w       <= '0';
                     sm_fifo_write_handler <= IDLE;
                  end if;

            end case;
         end if;
      end if;
   end process p_fifo_write_handler;

   p_fifo_read_handler : process(clk)

      variable v_log_line : line;                                                                             -- store log line

      file dataOut_file              : text;                                                                  -- data file
      variable v_dataOut_file_status : file_open_status;                                                      --file open status data file
      variable v_dataOut_line        : line;                                                                  --store single line from data file

      variable axi_word_cnt : integer := 0;

      variable temp_response : std_logic_vector(1 downto 0) := "00";
   begin
      if (rising_edge(clk)) then
         if rst_n = '0' then
            datafileID_read  <= (others => (others => '0'));
            length_read      <= (others => (others => '0'));
            words_in_fifo    <= (others => (others => '0'));
            fifo_cmd_ready_r <= (others => '0');
            handling_cmd_r   <= (others => '0');
         else

            case (sm_fifo_read_handler) is

               when IDLE =>
                  fifo_cmd_ready_r(i_r) <= '0';
                  if (i_r < NO_FIFOS - 1) then
                     i_r <= i_r + 1;
                  else
                     i_r <= 0;
                  end if;
                  sm_fifo_read_handler  <= HANDLE_FIFO_CMD;

               when HANDLE_FIFO_CMD =>

                  if (fifo_eof(i_r) /= '1' and fifo_waiting_on_sync(i_r) /= '1') then
                     if (fifo_cmd_valid(i_r) = '1' and fifo_cmd(i_r)(3) = '0') then
                        if (handling_cmd_r(i_r) = '0') then
                           addr_read(i_r) <= fifo_cmd(i_r)(35 downto 4);
                           if (fifo_cmd(i_r)(3 downto 0) = c_read_conf_no_check) then
                              bfm_read_req_r       <= '1';
                              sm_fifo_read_handler <= INITIATE_READ_CMD;

                           elsif (fifo_cmd(i_r)(3 downto 0) = c_read_conf_check_result) then
                              expected_data(i_r)   <= fifo_cmd(i_r)(67 downto 36);
                              bfm_read_req_r       <= '1';
                              sm_fifo_read_handler <= INITIATE_READ_CMD;

                           elsif (fifo_cmd(i_r)(3 downto 0) = c_read_data) then
                              datafileID_read(i_r) <= fifo_cmd(i_r)(67 downto 36);
                              length_read(i_r)     <= fifo_cmd(i_r)(99 downto 68);
                              bfm_read_req_r       <= '1';

                              handling_cmd_r(i_r) <= '1';

                              sm_fifo_read_handler <= INITIATE_STATUS_READ;
                           else
                              sm_fifo_read_handler <= IDLE;
                           end if;
                        else
                           bfm_read_req_r       <= '1';
                           sm_fifo_read_handler <= INITIATE_STATUS_READ;
                        end if;
                     else
                        sm_fifo_read_handler <= IDLE;
                     end if;
                  else
                     sm_fifo_read_handler <= IDLE;
                  end if;

               when INITIATE_READ_CMD =>

                  if (fifo_cmd(i_r)(3 downto 0) = c_read_conf_no_check) then
                     if (bfm_read_busy /= '1' and bfm_read_token = c_process_read) then
                        bfm_read_cmd_r(3 downto 0)  <= c_read_conf_no_check;
                        bfm_read_cmd_r(35 downto 4) <= addr_read(i_r);
                        bfm_read_cmd_valid_r        <= '1';

                        sm_fifo_read_handler <= WAIT_FOR_BFM_READ;
                     end if;

                  elsif (fifo_cmd(i_r)(3 downto 0) = c_read_conf_check_result) then
                     if (bfm_read_busy /= '1' and bfm_read_token = c_process_read) then
                        bfm_read_cmd_r(3 downto 0)  <= c_read_conf_no_check;
                        bfm_read_cmd_r(35 downto 4) <= addr_read(i_r);
                        bfm_read_cmd_valid_r        <= '1';

                        sm_fifo_read_handler <= WAIT_FOR_BFM_READ;
                     end if;

                  elsif (fifo_cmd(i_r)(3 downto 0) = c_read_data) then
                     if (bfm_read_busy /= '1' and bfm_read_token = c_process_read) then
                        if (length_read(i_r) /= 0) then

                           bfm_read_cmd_r(3 downto 0)   <= c_read_data;
                           bfm_read_cmd_r(35 downto 4)  <= addr_read(i_r);
                           bfm_read_cmd_r(67 downto 36) <= datafileID_read(i_r);
                           if (length_read(i_r) >= X"100") then
                              if (words_in_fifo(i_r) >= X"100") then
                                 bfm_read_cmd_r(99 downto 68) <= X"00000100";
                              else
                                 bfm_read_cmd_r(99 downto 68) <= X"00000" & words_in_fifo(i_r);
                              end if;
                           else
                              if (words_in_fifo(i_r) >= length_read(i_r)) then
                                 bfm_read_cmd_r(99 downto 68) <= length_read(i_r);
                              else
                                 bfm_read_cmd_r(99 downto 68) <= X"00000" & words_in_fifo(i_r);
                              end if;
                           end if;
                           bfm_read_cmd_valid_r         <= '1';

                           sm_fifo_read_handler <= WAIT_FOR_BFM_READ;
                        else
                           fifo_cmd_ready_r(i_r) <= '1';
                           sm_fifo_read_handler  <= IDLE;
                        end if;
                     end if;
                  end if;

               when WAIT_FOR_BFM_READ =>
                  bfm_read_cmd_valid_r <= '0';
                  if (bfm_read_done = true) then
                     length_read(i_r) <= length_read(i_r) - bfm_read_cmd_r(99 downto 68);

                     bfm_read_req_r       <= '0';
                     sm_fifo_read_handler <= LOG_DATA;
                  end if;

               when LOG_DATA =>

                  if (fifo_cmd(i_r)(3 downto 0) = c_read_conf_no_check) then

                     write(v_log_line, string'("Fifo") & integer'image(i_r) & string'(" Read from address 0x"));
                     write(v_log_line, to_hstring(fifo_cmd(i_r)(35 downto 4)));
                     write(v_log_line, string'(": 0x") & to_hstring(RD_DATA_from_master(31 downto 0)));
                     case (VRESPONSE(1 downto 0)) is
                        when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                        when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                        when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                        when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                        when others => write(v_log_line, string'(", AXI:Undefined Response"));
                     end case;
                     writeline(log_file, v_log_line);

                     fifo_cmd_ready_r(i_r) <= '1';
                     sm_fifo_read_handler  <= IDLE;

                  elsif (fifo_cmd(i_r)(3 downto 0) = c_read_conf_check_result) then

                     if (fifo_cmd(i_r)(67 downto 36) = RD_DATA_from_master(31 downto 0)) then
                        write(v_log_line, string'("Fifo") & integer'image(i_r) & string'(" Read from address 0x"));
                        write(v_log_line, to_hstring(fifo_cmd(i_r)(35 downto 4)));
                        write(v_log_line, string'(": 0x") & to_hstring(RD_DATA_from_master(31 downto 0)));
                        write(v_log_line, string'(", ok"));
                        report "*** Data read from register OK";
                     else
                        write(v_log_line, string'("ERROR: "));
                        write(v_log_line, string'("Fifo") & integer'image(i_r) & string'(" Read from address 0x"));
                        write(v_log_line, to_hstring(fifo_cmd(i_r)(35 downto 4)));
                        write(v_log_line, string'(": 0x") & to_hstring(RD_DATA_from_master(31 downto 0)));
                        write(v_log_line, string'(", expected "));
                        write(v_log_line, string'("0x") & to_hstring(fifo_cmd(i_r)(67 downto 36)));
                        report "*** Data read from register NOK!";
                     end if;

                     case (VRESPONSE(1 downto 0)) is
                        when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                        when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                        when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                        when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                        when others => write(v_log_line, string'(", AXI:Undefined Response"));
                     end case;
                     writeline(log_file, v_log_line);

                     fifo_cmd_ready_r(i_r) <= '1';
                     sm_fifo_read_handler  <= IDLE;

                  elsif (fifo_cmd(i_r)(3 downto 0) = c_read_data) then

                     file_open(v_dataOut_file_status, dataOut_file, LOG_FILE_DIR & (string'("/data/read/fifo") & integer'image(i_r) & string'("_") & integer'image(conv_integer(fifo_cmd(i_r)(67 downto 36))) & string'("_actual.txt")), append_mode);

                     if not (v_dataOut_file_status = OPEN_OK) then
                        write(v_log_line, string'(" ERROR: Unable to open resultdata file"));
                        writeline(log_file, v_log_line);
                        report "ERROR: Unable to open resultdata file"
                        severity FAILURE;
                     end if;

                     axi_word_cnt := 0;

                     --write words in hex to result file
                     for i in conv_integer(bfm_read_cmd_r(99 downto 68) - 1) downto 0 loop
                        axi_word_cnt := axi_word_cnt + 32;
                        write(v_dataOut_line, string'("0x") & to_hstring(RD_DATA_from_master(integer(axi_word_cnt) - 1 downto (integer(axi_word_cnt) - 32))));
                        writeline(dataOut_file, v_dataOut_line);
                     end loop;

                     file_close(dataOut_file);

                     write(v_log_line, string'("Fifo") & integer'image(i_r) & string'(" Read burst from address 0x"));
                     write(v_log_line, to_hstring(bfm_read_cmd_r(35 downto 4)));
                     write(v_log_line, string'(": ID ") & integer'image(conv_integer(bfm_read_cmd_r(67 downto 36))));
                     write(v_log_line, string'(": length: 0x") & to_hstring(bfm_read_cmd_r(99 downto 68)) & string'(" words"));

                     for i in 0 to conv_integer(bfm_read_cmd_r(99 downto 68) - 1) loop
                        case (VRESPONSE(((i * 2) + 1) downto (i * 2))) is
                           when OKAY   => temp_response := OKAY;
                           when EXOKAY => temp_response := EXOKAY;
                           when SLVERR =>
                              temp_response := SLVERR;
                              exit;
                           when DECERR =>
                              temp_response := DECERR;
                              exit;
                           when others =>
                              temp_response := "XX";
                              exit;
                        end case;
                     end loop;

                     case (temp_response) is
                        when OKAY   => write(v_log_line, string'(", AXI:OKAY"));
                        when EXOKAY => write(v_log_line, string'(", AXI:EXOKAY"));
                        when SLVERR => write(v_log_line, string'(", AXI:SLVERR"));
                        when DECERR => write(v_log_line, string'(", AXI:DECERR"));
                        when others => write(v_log_line, string'(", AXI:Undefined Response"));
                     end case;
                     writeline(log_file, v_log_line);

                     --              if(length_read(i_r) > 0) then
                     --                bfm_read_req_r          <= '1';
                     --                sm_fifo_read_handler    <= INITIATE_STATUS_READ;
                     --              else
                     --                fifo_cmd_ready_r(i_r)   <= '1';
                     --                handling_cmd_r(i_r)     <= '0';
                     --                sm_fifo_read_handler    <= IDLE;
                     --              end if;

                     if (length_read(i_r) > 0) then
                     --bfm_read_req_r          <= '1';
                     --sm_fifo_read_handler    <= INITIATE_STATUS_READ;
                     else
                        fifo_cmd_ready_r(i_r) <= '1';
                        handling_cmd_r(i_r)   <= '0';
                        --sm_fifo_read_handler    <= IDLE;
                     end if;

                     sm_fifo_read_handler <= IDLE;
                  end if;

               when INITIATE_STATUS_READ =>
                  if (bfm_read_busy /= '1' and bfm_read_token = c_process_read) then
                     bfm_read_cmd_r(3 downto 0)  <= c_req_read_fifo_status;
                     bfm_read_cmd_r(35 downto 4) <= std_logic_vector(to_unsigned((i_r - (NO_FIFOS / 2)), 32)); -- i_r - # of write fifos
                     bfm_read_cmd_valid_r        <= '1';

                     sm_fifo_read_handler <= WAIT_FOR_STATUS_READ;
                  end if;

               when WAIT_FOR_STATUS_READ =>
                  bfm_read_cmd_valid_r <= '0';
                  if (bfm_read_done = true) then
                     if (RD_DATA_from_master(11 downto 0) = X"000") then
                        bfm_write_req_r      <= '1';
                        sm_fifo_read_handler <= SET_THRESHOLD;
                     else
                        words_in_fifo(i_r)   <= RD_DATA_from_master(11 downto 0);
                        bfm_read_req_r       <= '1';
                        sm_fifo_read_handler <= INITIATE_READ_CMD;
                     end if;

                     bfm_read_req_r <= '0';
                  end if;

               when SET_THRESHOLD =>
                  if (bfm_write_busy /= '1' and bfm_write_token = c_process_read) then
                     bfm_write_cmd_r(3 downto 0)  <= c_set_read_threshold;
                     bfm_write_cmd_r(35 downto 4) <= std_logic_vector(to_unsigned((i_r - (NO_FIFOS / 2)), 32)); -- i_r - # of write fifos
                     if (length_read(i_r) >= (c_cpu_read_fifo_size / 2)) then
                        bfm_write_cmd_r(67 downto 36) <= std_logic_vector(to_unsigned(c_cpu_read_fifo_size / 2, 32));
                     else
                        bfm_write_cmd_r(67 downto 36) <= length_read(i_r);
                     end if;
                     bfm_write_cmd_valid_r        <= '1';

                     sm_fifo_read_handler <= WAIT_FOR_THRESHOLD_SET;
                  end if;

               when WAIT_FOR_THRESHOLD_SET =>
                  bfm_write_cmd_valid_r <= '0';
                  if (bfm_write_done = true) then
                     bfm_write_cmd_r(3 downto 0)  <= c_set_read_irq;
                     bfm_write_cmd_r(35 downto 4) <= std_logic_vector(to_unsigned((i_r - (NO_FIFOS / 2)), 32)); -- i_r - # of write fifos
                     bfm_write_cmd_valid_r        <= '1';

                     sm_fifo_read_handler <= WAIT_FOR_IRQ_SET;
                  end if;

               when WAIT_FOR_IRQ_SET =>
                  bfm_write_cmd_valid_r <= '0';
                  if (bfm_write_done = true) then
                     bfm_write_req_r      <= '0';
                     sm_fifo_read_handler <= IDLE;
                  end if;

            end case;
         end if;
      end if;
   end process p_fifo_read_handler;

   fifo_cmd_ready <= fifo_cmd_ready_w or fifo_cmd_ready_r;


   p_drive_write_bfm : process(clk)

      variable v_log_line : line;                                                                             -- store log line

      file dataIn_file              : text;                                                                   -- data file
      variable v_dataIn_file_status : file_open_status;                                                       --file open status data file

      variable axi_data     : std_logic_vector(8191 downto 0) := (others => '0');
      variable axi_word     : string(1 to 10);
      variable axi_word_cnt : integer                         := 0;

      variable line_cnt : integer := 0;

      variable stream_id     : integer := 0;
      variable node_offset   : integer := 0;
      variable stream_offset : integer := 0;
      variable addr          : integer := 0;
   begin
      if (rising_edge(clk)) then
         if rst_n = '0' then

            wr_cmd_valid <= '0';

            sm_bfm_write <= WAIT_FOR_CMD;
         else
            case (sm_bfm_write) is

               when WAIT_FOR_CMD =>
                  bfm_write_done <= false;

                  if (bfm_write_cmd_valid = '1') then
                     bfm_write_busy <= '1';
                     if (bfm_write_cmd(3 downto 0) = c_write_conf) then                                       -- set AXI signals

                        wr_cmd_address           <= bfm_write_cmd(35 downto 4);
                        wr_cmd_length            <= (others => '0');
                        wr_cmd_data(31 downto 0) <= bfm_write_cmd(67 downto 36);

                        wr_cmd_valid <= '1';                                                                  -- start write task

                        sm_bfm_write <= WAIT_FOR_BFM_WRITE;

                     elsif (bfm_write_cmd(3 downto 0) = c_write_data) then                                    -- set AXI signals

                        file_open(v_dataIn_file_status, dataIn_file, (LOG_FILE_DIR & string'("/data/write/fifo") & integer'image(i_w) & string'("_") & integer'image(conv_integer(bfm_write_cmd(67 downto 36))) & string'(".txt")), read_mode);

                        if not (v_dataIn_file_status = OPEN_OK) then
                           write(v_log_line, string'(" ERROR: Unable to open data file"));
                           writeline(log_file, v_log_line);
                           report "ERROR: Unable to open data file"
                           severity FAILURE;
                        end if;

                        line_cnt     := 0;
                        axi_word_cnt := 0;
                        axi_data     := (others => '0');

                        --read lines from data file
                        while (not endfile(dataIn_file)) and (axi_word_cnt < 8192) loop
                           str_read(dataIn_file, axi_word);
                           line_cnt := line_cnt + 1;
                           if (line_cnt > words_send(i_w)) then
                              axi_word_cnt                                                            := axi_word_cnt + 32;
                              axi_data(integer(axi_word_cnt) - 1 downto (integer(axi_word_cnt) - 32)) := hstr_to_slv(axi_word(3 to 10));
                           end if;
                        end loop;

                        file_close(dataIn_file);

                        wr_cmd_address              <= bfm_write_cmd(35 downto 4);
                        wr_cmd_length(7 downto 0)   <= bfm_write_cmd(75 downto 68) - 1;
                        wr_cmd_data(axi_data'range) <= axi_data;

                        wr_cmd_valid <= '1';                                                                  -- start write task

                        sm_bfm_write <= WAIT_FOR_BFM_WRITE;

                     elsif (bfm_write_cmd(3 downto 0) = c_clear_write_irq) then

                        wr_cmd_address           <= c_irq_clear_reg_w_addr;
                        wr_cmd_length            <= (others => '0');
                        wr_cmd_data(31 downto 0) <= bfm_write_cmd(35 downto 4);

                        wr_cmd_valid <= '1';                                                                  -- start write task

                        sm_bfm_write <= WAIT_FOR_BFM_WRITE;

                     elsif (bfm_write_cmd(3 downto 0) = c_clear_read_irq) then

                        wr_cmd_address           <= c_irq_clear_reg_r_addr;
                        wr_cmd_length            <= (others => '0');
                        wr_cmd_data(31 downto 0) <= bfm_write_cmd(35 downto 4);

                        wr_cmd_valid <= '1';                                                                  -- start write task

                        sm_bfm_write <= WAIT_FOR_BFM_WRITE;

                     elsif (bfm_write_cmd(3 downto 0) = c_set_read_threshold) then

                        stream_id     := conv_integer(bfm_read_cmd(35 downto 4));
                        node_offset   := integer(floor(real(stream_id / 4)));
                        stream_offset := stream_id - (node_offset * 4);
                        addr          := conv_integer(X"64400800") + (node_offset * conv_integer(X"800")) + conv_integer(X"420") + (stream_offset * 4);

                        wr_cmd_address           <= std_logic_vector(to_unsigned(addr, 32));
                        wr_cmd_length            <= (others => '0');
                        wr_cmd_data(31 downto 0) <= bfm_write_cmd(67 downto 36);

                        wr_cmd_valid <= '1';                                                                  -- start write task

                        sm_bfm_write <= WAIT_FOR_BFM_WRITE;

                     elsif (bfm_write_cmd(3 downto 0) = c_set_write_threshold) then

                        stream_id     := conv_integer(bfm_read_cmd(35 downto 4));
                        node_offset   := integer(floor(real(stream_id / 4)));
                        stream_offset := stream_id - (node_offset * 4);
                        addr          := conv_integer(X"64400800") + (node_offset * conv_integer(X"800")) + conv_integer(X"520") + (stream_offset * 4);

                        wr_cmd_address           <= std_logic_vector(to_unsigned(addr, 32));
                        wr_cmd_length            <= (others => '0');
                        wr_cmd_data(31 downto 0) <= bfm_write_cmd(67 downto 36);

                        wr_cmd_valid <= '1';                                                                  -- start write task

                        sm_bfm_write <= WAIT_FOR_BFM_WRITE;

                     elsif (bfm_write_cmd(3 downto 0) = c_set_read_irq) then

                        wr_cmd_address           <= c_irq_set_reg_r_addr;
                        wr_cmd_length            <= (others => '0');
                        wr_cmd_data(31 downto 0) <= (conv_integer(bfm_write_cmd(35 downto 4)) => '1', others => '0');

                        wr_cmd_valid <= '1';                                                                  -- start write task

                        sm_bfm_write <= WAIT_FOR_BFM_WRITE;

                     elsif (bfm_write_cmd(3 downto 0) = c_set_write_irq) then

                        wr_cmd_address           <= c_irq_set_reg_w_addr;
                        wr_cmd_length            <= (others => '0');
                        wr_cmd_data(31 downto 0) <= (conv_integer(bfm_write_cmd(35 downto 4)) => '1', others => '0');

                        wr_cmd_valid <= '1';                                                                  -- start write task

                        sm_bfm_write <= WAIT_FOR_BFM_WRITE;

                     end if;
                  end if;

               when WAIT_FOR_BFM_WRITE =>

                  if (wr_cmd_done = '1') then
                     wr_cmd_valid   <= '0';
                     bfm_write_done <= true;
                     bfm_write_busy <= '0';
                     sm_bfm_write   <= WAIT_FOR_CMD;
                  end if;

            end case;
         end if;
      end if;
   end process p_drive_write_bfm;

   p_drive_read_bfm : process(clk)

      variable v_log_line : line;                                                                             -- store log line

      variable axi_data     : std_logic_vector(8191 downto 0) := (others => '0');
      variable axi_word     : string(1 to 10);
      variable axi_word_cnt : integer                         := 0;

      variable stream_id     : integer := 0;
      variable node_offset   : integer := 0;
      variable stream_offset : integer := 0;
      variable addr          : integer := 0;

   begin
      if (rising_edge(clk)) then
         if (rst_n = '0') then
            bfm_read_done <= false;
            rd_cmd_valid  <= '0';

            sm_bfm_read <= WAIT_FOR_CMD;
         else
            case (sm_bfm_read) is

               when WAIT_FOR_CMD =>
                  bfm_read_done <= false;

                  if (bfm_read_cmd_valid = '1') then
                     bfm_read_busy <= '1';

                     if (bfm_read_cmd(3 downto 0) = c_read_conf_no_check) then                                -- set AXI signals

                        rd_cmd_address <= bfm_read_cmd(35 downto 4);
                        rd_cmd_length  <= (others => '0');

                        rd_cmd_valid <= '1';                                                                  -- start read task

                        sm_bfm_read <= WAIT_FOR_BFM_READ;

                     elsif (bfm_read_cmd(3 downto 0) = c_read_conf_check_result) then                         -- set AXI signals

                        rd_cmd_address <= bfm_read_cmd(35 downto 4);
                        rd_cmd_length  <= (others => '0');

                        rd_cmd_valid <= '1';                                                                  -- start read task

                        sm_bfm_read <= WAIT_FOR_BFM_READ;

                     elsif (bfm_read_cmd(3 downto 0) = c_read_data) then                                      -- set AXI signals

                        rd_cmd_address            <= bfm_read_cmd(35 downto 4);
                        rd_cmd_length(7 downto 0) <= bfm_read_cmd(75 downto 68) - 1;

                        rd_cmd_valid <= '1';                                                                  -- start read task

                        sm_bfm_read <= WAIT_FOR_BFM_READ;

                     elsif (bfm_read_cmd(3 downto 0) = c_req_irq_info) then

                        rd_cmd_address            <= c_irq_status_regs_addr;
                        rd_cmd_length(7 downto 0) <= X"04";                                                   -- to read from 824 to 834

                        rd_cmd_valid <= '1';                                                                  -- start read task

                        sm_bfm_read <= WAIT_FOR_BFM_READ;

                     elsif (bfm_read_cmd(3 downto 0) = c_req_read_fifo_status) then
                        stream_id     := conv_integer(bfm_read_cmd(35 downto 4));
                        node_offset   := integer(floor(real(stream_id / 4)));
                        stream_offset := stream_id - (node_offset * 4);
                        addr          := conv_integer(X"64400800") + (node_offset * conv_integer(X"800")) + conv_integer(X"400") + (stream_offset * 4);

                        rd_cmd_address <= std_logic_vector(to_unsigned(addr, 32));
                        rd_cmd_length  <= (others => '0');

                        rd_cmd_valid <= '1';                                                                  -- start read task

                        sm_bfm_read <= WAIT_FOR_BFM_READ;

                     elsif (bfm_read_cmd(3 downto 0) = c_req_write_fifo_status) then
                        stream_id     := conv_integer(bfm_read_cmd(35 downto 4));
                        node_offset   := integer(floor(real(stream_id / 4)));
                        stream_offset := stream_id - (node_offset * 4);
                        addr          := conv_integer(X"64400800") + (node_offset * conv_integer(X"800")) + conv_integer(X"500") + (stream_offset * 4);

                        rd_cmd_address <= std_logic_vector(to_unsigned(addr, 32));
                        rd_cmd_length  <= (others => '0');

                        rd_cmd_valid <= '1';                                                                  -- start read task

                        sm_bfm_read <= WAIT_FOR_BFM_READ;

                     end if;
                  end if;

               when WAIT_FOR_BFM_READ =>

                  if (rd_cmd_done = '1') then
                     RD_DATA_from_master <= rd_cmd_data(RD_DATA_from_master'range);
                     rd_cmd_valid        <= '0';
                     bfm_read_done       <= true;
                     bfm_read_busy       <= '0';
                     sm_bfm_read         <= WAIT_FOR_CMD;
                  end if;

            end case;
         end if;
      end if;
   end process p_drive_read_bfm;

   p_bfm_write_arbiter : process(clk)
   begin
      if (rising_edge(clk)) then
         if rst_n = '0' then
            bfm_write_token <= c_process_write;
         else
            if (arbitter = c_fifos) then
               if (switch_write_token = '1') then
                  if (handle_irq = '1') then
                     switch_write_token <= '0';
                     bfm_write_token    <= c_process_irq;
                  elsif (bfm_write_req_r = '1') then
                     switch_write_token <= '0';
                     bfm_write_token    <= c_process_read;
                  elsif (bfm_write_req_w = '1') then
                     switch_write_token <= '0';
                     bfm_write_token    <= c_process_write;
                  else
                     bfm_write_token <= c_no_process;
                  end if;
               else
                  if (bfm_write_token = c_process_irq) then
                     if (handle_irq = '0') then
                        switch_write_token <= '1';
                     end if;
                  elsif (bfm_write_token = c_process_read) then
                     if (bfm_write_req_r = '0') then
                        switch_write_token <= '1';
                     end if;
                  elsif (bfm_write_token = c_process_write) then
                     if (bfm_write_req_w = '0') then
                        switch_write_token <= '1';
                     end if;
                  else
                     switch_write_token <= '1';
                  end if;
               end if;
            else
               bfm_write_token <= c_process_nodes;
            end if;
         end if;
      end if;

      if (bfm_write_token = c_process_nodes) then
         bfm_write_cmd       <= bfm_write_cmd_nodes;
         bfm_write_cmd_valid <= bfm_write_cmd_valid_nodes;
      else
         if (bfm_write_token = c_process_irq) then
            bfm_write_cmd       <= bfm_write_cmd_i;
            bfm_write_cmd_valid <= bfm_write_cmd_valid_i;
         elsif (bfm_write_token = c_process_write) then
            bfm_write_cmd       <= bfm_write_cmd_w;
            bfm_write_cmd_valid <= bfm_write_cmd_valid_w;
         elsif (bfm_write_token = c_process_read) then
            bfm_write_cmd       <= bfm_write_cmd_r;
            bfm_write_cmd_valid <= bfm_write_cmd_valid_r;
         else
            bfm_write_cmd       <= (others => '0');
            bfm_write_cmd_valid <= '0';
         end if;
      end if;

   end process p_bfm_write_arbiter;

   p_bfm_read_arbiter : process(clk)
   begin
      if (rising_edge(clk)) then
         if rst_n = '0' then
            bfm_read_token <= c_process_write;
         else
            if (arbitter = c_fifos) then
               if (switch_read_token = '1') then
                  if (handle_irq = '1') then
                     switch_read_token <= '0';
                     bfm_read_token    <= c_process_irq;
                  else
                     if (bfm_read_req_w = '1') then
                        switch_read_token <= '0';
                        bfm_read_token    <= c_process_write;
                     elsif (bfm_read_req_r = '1') then
                        switch_read_token <= '0';
                        bfm_read_token    <= c_process_read;
                     else
                        bfm_read_token <= c_no_process;
                     end if;
                  end if;

               else
                  if (bfm_read_token = c_process_irq) then
                     if (handle_irq = '0') then
                        switch_read_token <= '1';
                     end if;
                  elsif (bfm_read_token = c_process_read) then
                     if (bfm_read_req_r = '0') then
                        switch_read_token <= '1';
                     end if;
                  elsif (bfm_read_token = c_process_write) then
                     if (bfm_read_req_w = '0') then
                        switch_read_token <= '1';
                     end if;
                  else
                     switch_read_token <= '1';
                  end if;
               end if;
            else
               bfm_read_token <= c_process_nodes;
            end if;
         end if;
      end if;

      if (bfm_read_token = c_process_nodes) then
         bfm_read_cmd       <= bfm_read_cmd_nodes;
         bfm_read_cmd_valid <= bfm_read_cmd_valid_nodes;
      else
         if (bfm_read_token = c_process_irq) then
            bfm_read_cmd       <= bfm_read_cmd_i;
            bfm_read_cmd_valid <= bfm_read_cmd_valid_i;
         elsif (bfm_read_token = c_process_write) then
            bfm_read_cmd       <= bfm_read_cmd_w;
            bfm_read_cmd_valid <= bfm_read_cmd_valid_w;
         elsif (bfm_read_token = c_process_read) then
            bfm_read_cmd       <= bfm_read_cmd_r;
            bfm_read_cmd_valid <= bfm_read_cmd_valid_r;
         else
            bfm_read_cmd       <= (others => '0');
            bfm_read_cmd_valid <= '0';
         end if;
      end if;

   end process p_bfm_read_arbiter;

   p_syncer : process(clk)
   begin
      if (rising_edge(clk)) then
         if rst_n = '0' then
            --State to start control/cpu/hdl/fifos in that order
            arbitter  <= c_control_node;
            sm_syncer <= WAITING_FOR_SYNC;
            --State to start control/cpu/hdl/fifos in that order

            control_continue_after_sync <= '0';
            cpu_continue_after_sync     <= '0';
            hdl_continue_after_sync     <= '0';
            fifo_continue_after_sync    <= (others => '0');
         else
            case (sm_syncer) is
               when WAITING_FOR_SYNC =>
                  if (sync_delay /= '0') then
                     sync_delay               <= '0';
                     fifo_continue_after_sync <= (others => '0');
                  else
                     if ((fifo_waiting_on_sync or fifo_eof) = (NO_FIFOS - 1 downto 0 => '1') and
                              control_waiting_on_sync = '1' and
                              cpu_waiting_on_sync = '1' and
                              hdl_waiting_on_sync = '1') then

                        write(v_log_line, string'("Continue after sync"));
                        writeline(log_file, v_log_line);

                        arbitter                    <= c_control_node;
                        control_continue_after_sync <= '1';
                        sync_delay                  <= '1';
                        sm_syncer                   <= WAITING_FOR_CTRL;
                     end if;
                  end if;
               when WAITING_FOR_CTRL =>
                  if (sync_delay /= '0') then
                     sync_delay                  <= '0';
                     control_continue_after_sync <= '0';
                  else
                     if (control_waiting_on_sync = '1' or control_eof = '1') then
                        arbitter                <= c_cpu_node;
                        cpu_continue_after_sync <= '1';
                        sync_delay              <= '1';
                        sm_syncer               <= WAITING_FOR_CPU;
                     end if;
                  end if;
               when WAITING_FOR_CPU =>
                  if (sync_delay /= '0') then
                     sync_delay              <= '0';
                     cpu_continue_after_sync <= '0';
                  else
                     if (cpu_waiting_on_sync = '1' or cpu_eof = '1') then
                        arbitter                <= c_hdl_node;
                        hdl_continue_after_sync <= '1';
                        sync_delay              <= '1';
                        sm_syncer               <= WAITING_FOR_HDL;
                     end if;
                  end if;

               when WAITING_FOR_HDL =>
                  if (sync_delay /= '0') then
                     sync_delay              <= '0';
                     hdl_continue_after_sync <= '0';
                  else
                     if (hdl_waiting_on_sync = '1' or hdl_eof = '1') then
                        arbitter                 <= c_fifos;
                        fifo_continue_after_sync <= (others => '1');
                        sync_delay               <= '1';
                        sm_syncer                <= WAITING_FOR_SYNC;
                     end if;
                  end if;
            end case;
         end if;
      end if;
   end process p_syncer;

   p_eof : process(clk)
   begin
      if (rising_edge(clk)) then
         if rst_n = '0' then
         else
            if (fifo_eof = (NO_FIFOS - 1 downto 0 => '1') and
                control_eof = '1' and
                cpu_eof = '1' and
                hdl_eof = '1') then
               file_close(log_file);
               report "*** End of simulation ***";
               finish(0);
            end if;
         end if;
      end if;
   end process;

end behaviour;
